# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import logging
import os
import re
import subprocess
import time

from tornado.ioloop import IOLoop

from motioneye import config, settings, utils


def start():
    io_loop = IOLoop.current()
    io_loop.add_timeout(
        datetime.timedelta(seconds=settings.MOUNT_CHECK_INTERVAL), _check_mounts
    )


def stop():
    _umount_all()


def find_mount_cifs():
    try:
        return utils.call_subprocess(['which', 'mount.cifs'])

    except subprocess.CalledProcessError:  # not found
        return None


def make_mount_point(server, share, username):
    server = re.sub('[^a-zA-Z0-9]', '_', server).lower()
    share = re.sub('[^a-zA-Z0-9]', '_', share).lower()

    if username:
        username = re.sub('[^a-zA-Z0-9]', '_', username).lower()
        mount_point = os.path.join(
            settings.SMB_MOUNT_ROOT, f'motioneye_{server}_{share}_{username}'
        )

    else:
        mount_point = os.path.join(
            settings.SMB_MOUNT_ROOT, f'motioneye_{server}_{share}'
        )

    return mount_point


def list_mounts():
    logging.debug('listing smb mounts...')

    mounts = []
    with open('/proc/mounts') as f:
        for line in f:
            line = line.strip()
            if not line:
                continue
            parts = line.split()
            if len(parts) < 4:
                continue

            target = parts[0]
            mount_point = parts[1]
            fstype = parts[2]
            opts = ' '.join(parts[3:])

            if fstype != 'cifs':
                continue

            if not _is_motioneye_mount(mount_point):
                continue

            match = re.match('//([^/]+)/(.+)', target)
            if not match:
                continue

            if len(match.groups()) != 2:
                continue

            server, share = match.groups()
            share = share.replace(
                '\\040', ' '
            )  # spaces are reported oddly by /proc/mounts

            match = re.search(r'username=([a-z][-\w]*)', opts)
            if match:
                username = match.group(1)

            else:
                username = ''

            match = re.search(r'vers=([\w.]+)', opts)
            if match:
                smb_ver = match.group(1)

            else:
                smb_ver = '1.0'

            logging.debug(f'found smb mount "//{server}/{share}" at "{mount_point}"')

            mounts.append(
                {
                    'server': server.lower(),
                    'share': share.lower(),
                    'smb_ver': smb_ver,
                    'username': username,
                    'mount_point': mount_point,
                }
            )

    return mounts


def update_mounts():
    network_shares = config.get_network_shares()

    mounts = list_mounts()
    mounts = {
        (m['server'], m['share'], m['smb_ver'], m['username'] or ''): False
        for m in mounts
    }

    should_stop = False  # indicates that motion should be stopped immediately
    should_start = True  # indicates that motion can be started afterwards
    for network_share in network_shares:
        key = (
            network_share['server'].lower(),
            network_share['share'].lower(),
            network_share['smb_ver'],
            network_share['username'].lower() or '',
        )

        if key in mounts:  # found
            mounts[key] = True

        else:  # needs to be mounted
            should_stop = True
            if not _mount(
                network_share['server'],
                network_share['share'],
                network_share['smb_ver'],
                network_share['username'],
                network_share['password'],
            ):
                should_start = False

    # unmount the no longer necessary mounts
    for (server, share, smb_ver, username), required in list(mounts.items()):
        if not required:
            _umount(server, share, username)
            should_stop = True

    return should_stop, should_start


def test_share(server, share, smb_ver, username, password, root_directory):
    mounts = list_mounts()
    mounts = {
        (m['server'], m['share'], m['smb_ver'], m['username'] or ''): m['mount_point']
        for m in mounts
    }

    key = (server, share, smb_ver, username or '')
    mounted = False
    mount_point = mounts.get(key)
    if not mount_point:
        mount_point = _mount(server, share, smb_ver, username, password)
        if not mount_point:
            raise Exception('cannot mount network share')

        mounted = True

    def maybe_umount():
        if mounted:
            time.sleep(1)
            _umount(server, share, username)

    path = os.path.join(mount_point, root_directory)
    if os.path.exists(path):
        return maybe_umount()

    try:
        os.makedirs(path)

    except:
        raise Exception('cannot create root directory')

    finally:
        maybe_umount()


def _mount(server, share, smb_ver, username, password):
    mount_point = make_mount_point(server, share, username)

    logging.debug('making sure mount point "%s" exists' % mount_point)

    if not os.path.exists(mount_point):
        os.makedirs(mount_point)

    if username:
        opts = f'username={username},password={password}'
        sec_types = [None, 'ntlm', 'ntlmv2', 'ntlmv2i', 'ntlmsspi', 'none']

    else:
        opts = 'guest'
        sec_types = [None, 'none', 'ntlm', 'ntlmv2', 'ntlmv2i', 'ntlmsspi']

    opts += ',vers=%s' % smb_ver

    for sec in sec_types:
        if sec:
            actual_opts = opts + ',sec=' + sec

        else:
            actual_opts = opts

        try:
            logging.debug(
                f'mounting "//{server}/{share}" at "{mount_point}" (sec={sec})'
            )
            subprocess.run(
                ['mount.cifs', f'//{server}/{share}', mount_point, '-o', actual_opts],
                check=True,
            )
            break

        except subprocess.CalledProcessError:
            pass

    else:
        logging.error(
            f'failed to mount smb share "//{server}/{share}" at "{mount_point}"'
        )
        return None

    # test to see if mount point is writable
    try:
        path = os.path.join(mount_point, '.motioneye_' + str(int(time.time())))
        os.mkdir(path)
        os.rmdir(path)
        logging.debug('directory at "%s" is writable' % mount_point)

    except:
        logging.error('directory at "%s" is not writable' % mount_point)

        return None

    return mount_point


def _umount(server, share, username):
    mount_point = make_mount_point(server, share, username)
    logging.debug(f'unmounting "//{server}/{share}" from "{mount_point}"')

    try:
        subprocess.run(['umount', mount_point], check=True)

    except subprocess.CalledProcessError:
        logging.error(
            f'failed to unmount smb share "//{server}/{share}" from "{mount_point}"'
        )

        return False

    try:
        os.rmdir(mount_point)

    except Exception as e:
        logging.error(f'failed to remove smb mount point "{mount_point}": {e}')

        return False

    return True


def _is_motioneye_mount(mount_point):
    mount_point_root = os.path.join(settings.SMB_MOUNT_ROOT, 'motioneye_')
    return bool(re.match('^' + mount_point_root + r'\w+$', mount_point))


def _umount_all():
    for mount in list_mounts():
        _umount(mount['server'], mount['share'], mount['username'])


def _check_mounts():
    from motioneye import motionctl

    logging.debug('checking SMB mounts...')

    stop, start = update_mounts()
    if stop:
        motionctl.stop()

    if start:
        motionctl.start()

    io_loop = IOLoop.current()
    io_loop.add_timeout(
        datetime.timedelta(seconds=settings.MOUNT_CHECK_INTERVAL), _check_mounts
    )



import inspect
from dataclasses import dataclass
from typing import Any, Hashable, Union

__all__ = ('RtmpUrl', 'RtspUrl', 'MjpegUrl')


@dataclass
class StreamUrl:
    scheme: str
    port: str
    host: str = '127.0.0.1'
    path: str = ''
    username: Union[None, str] = None
    password: Union[None, str] = None

    _tpl = '%(scheme)s://%(host)s%(port)s%(path)s'

    def __str__(self):
        return self._tpl % dict(
            scheme=self.scheme,
            host=self.host,
            port=(':' + str(self.port)) if self.port else '',
            path=self.path,
            username=self.username,
            password=self.password,
        )

    @classmethod
    def _get_dict_field_val(cls, k: Union[str, Hashable], v: Any) -> Any:
        try:
            return v or getattr(cls, k)
        except AttributeError:
            return v

    @classmethod
    def from_dict(cls, d: dict):  # -> typing.Self in Python >= v3.11
        return cls(
            **{
                k: cls._get_dict_field_val(k, v)
                for k, v in d.items()
                if k in inspect.signature(cls).parameters
            }
        )


@dataclass
class RtmpUrl(StreamUrl):
    scheme: str = 'rtmp'
    port: str = '1935'


@dataclass
class RtspUrl(StreamUrl):
    scheme: str = 'rtsp'
    port: str = '554'


@dataclass
class MjpegUrl(StreamUrl):
    scheme: str = 'http'
    port: str = '80'



# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import subprocess
from collections import OrderedDict
from typing import Dict

from motioneye import utils

__all__ = ('PowerControl',)


class PowerControl:
    _shut_down_cmd_sequence = OrderedDict(
        [
            ('poweroff', ''),
            ('shutdown', ' -h now'),
            ('systemctl', ' poweroff'),
            ('init', ' 0'),
        ]
    )

    _reboot_cmd_sequence = OrderedDict(
        [
            ('reboot', ''),
            ('shutdown', ' -r now'),
            ('systemctl', ' reboot'),
            ('init', ' 6'),
        ]
    )

    @staticmethod
    def _find_prog(prog: str) -> str:
        return utils.call_subprocess(['which', prog])

    @classmethod
    def _exec_prog(cls, prog: str, args: str = '') -> bool:
        p = cls._find_prog(prog)
        logging.info('executing "%s"' % p)
        return os.system(p + args) == 0

    @classmethod
    def _run_procedure(cls, prog_sequence: Dict[str, str], log_msg: str) -> bool:
        logging.info(log_msg)

        for prog, args in prog_sequence.items():
            try:
                return cls._exec_prog(prog, args)
            except subprocess.CalledProcessError:  # program not found
                continue
        else:
            return False

    @classmethod
    def shut_down(cls) -> bool:
        return cls._run_procedure(cls._shut_down_cmd_sequence, 'shutting down')

    @classmethod
    def reboot(cls) -> bool:
        return cls._run_procedure(cls._reboot_cmd_sequence, 'rebooting')



# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import re
from typing import Union

from tornado.httpclient import AsyncHTTPClient, HTTPRequest, HTTPResponse

from motioneye import settings, utils

_DOUBLE_SLASH_REGEX = re.compile('//+')


def _make_request(
    scheme,
    host,
    port,
    username,
    password,
    path,
    method='GET',
    data=None,
    query=None,
    timeout=None,
    content_type=None,
):
    path = _DOUBLE_SLASH_REGEX.sub('/', path)
    url = '{scheme}://{host}{port}{path}'.format(
        scheme=scheme, host=host, port=':' + str(port) if port else '', path=path or ''
    )

    query = dict(query or {})
    query['_username'] = username or ''
    query['_admin'] = 'true'  # always use the admin account

    if url.count('?'):
        url += '&'

    else:
        url += '?'

    url += '&'.join([(n + '=' + v) for (n, v) in list(query.items())])
    url += '&_signature=' + utils.compute_signature(method, url, data, password)

    if timeout is None:
        timeout = settings.REMOTE_REQUEST_TIMEOUT

    headers = {}
    if content_type:
        headers['Content-Type'] = content_type

    return HTTPRequest(
        url,
        method,
        body=data,
        connect_timeout=timeout,
        request_timeout=timeout,
        headers=headers,
        validate_cert=settings.VALIDATE_CERTS,
    )


async def _send_request(request: HTTPRequest) -> HTTPResponse:
    response = await AsyncHTTPClient().fetch(request, raise_error=False)

    if response.code != 200:
        try:
            decoded = json.loads(response.body)
            if decoded['error'] == 'unauthorized':
                response.error = 'Authentication Error'

            elif decoded['error']:
                response.error = decoded['error']

        except Exception as e:
            logging.error(f"_send_request: {e}")

    return response


def pretty_camera_url(local_config, camera=True):
    scheme = local_config.get('@scheme', local_config.get('scheme')) or 'http'
    host = local_config.get('@host', local_config.get('host'))
    port = local_config.get('@port', local_config.get('port'))
    path = local_config.get('@path', local_config.get('path')) or ''

    url = scheme + '://' + host
    if port and str(port) not in ['80', '443']:
        url += ':' + str(port)

    if path:
        url += path

    if url.endswith('/'):
        url = url[:-1]

    if camera:
        if camera is True:
            url += '/config/' + str(
                local_config.get(
                    '@remote_camera_id', local_config.get('remote_camera_id')
                )
            )

        else:
            url += '/config/' + str(camera)

    return url


def _remote_params(local_config):
    params = [
        local_config.get('@scheme', local_config.get('scheme')) or 'http',
        local_config.get('@host', local_config.get('host')),
        local_config.get('@port', local_config.get('port')),
        local_config.get('@username', local_config.get('username')),
        local_config.get('@password', local_config.get('password')),
        local_config.get('@path', local_config.get('path')) or '',
        local_config.get('@remote_camera_id', local_config.get('remote_camera_id')),
    ]

    if params[3] is not None:
        params[3] = str(params[3])

    if params[4] is not None:
        params[4] = str(params[4])

    return params


def make_camera_response(c):
    return {'id': c['id'], 'name': c['name']}


async def list_cameras(local_config) -> utils.GetCamerasResponse:
    scheme, host, port, username, password, path, _ = _remote_params(local_config)

    logging.debug(
        'listing remote cameras on {url}'.format(
            url=pretty_camera_url(local_config, camera=False)
        )
    )

    request = _make_request(
        scheme, host, port, username, password, path + '/config/list/'
    )

    response = await _send_request(request)

    if response.error:
        logging.error(
            'failed to list remote cameras on {url}: {msg}'.format(
                url=pretty_camera_url(local_config, camera=False),
                msg=utils.pretty_http_error(response),
            )
        )

        return utils.GetCamerasResponse(None, utils.pretty_http_error(response))

    try:
        response = json.loads(response.body)

    except Exception as e:
        logging.error(
            'failed to decode json answer from {url}: {msg}'.format(
                url=pretty_camera_url(local_config, camera=False), msg=str(e)
            )
        )

        return utils.GetCamerasResponse(None, str(e))

    else:
        cameras = response['cameras']

        # filter out simple mjpeg cameras
        cameras = [
            make_camera_response(c)
            for c in cameras
            if c['proto'] != 'mjpeg' and c.get('enabled')
        ]

        return utils.GetCamerasResponse(cameras, None)


async def get_config(local_config) -> utils.GetConfigResponse:
    scheme, host, port, username, password, path, camera_id = _remote_params(
        local_config
    )

    logging.debug(
        'getting config for remote camera {id} on {url}'.format(
            id=camera_id, url=pretty_camera_url(local_config)
        )
    )

    request = _make_request(
        scheme, host, port, username, password, path + f'/config/{camera_id}/get/'
    )
    response = await _send_request(request)

    if response.error:
        logging.error(
            'failed to get config for remote camera {id} on {url}: {msg}'.format(
                id=camera_id,
                url=pretty_camera_url(local_config),
                msg=utils.pretty_http_error(response),
            )
        )

        return utils.GetConfigResponse(None, error=utils.pretty_http_error(response))

    try:
        response = json.loads(response.body)

    except Exception as e:
        logging.error(
            'failed to decode json answer from {url}: {msg}'.format(
                url=pretty_camera_url(local_config), msg=str(e)
            )
        )

        return utils.GetConfigResponse(None, error=str(e))

    else:
        response['host'] = host
        response['port'] = port

        return utils.GetConfigResponse(remote_ui_config=response, error=None)


async def set_config(local_config, ui_config) -> Union[str, None]:
    scheme = local_config.get('@scheme', local_config.get('scheme'))
    host = local_config.get('@host', local_config.get('host'))
    port = local_config.get('@port', local_config.get('port'))
    username = local_config.get('@username', local_config.get('username'))
    password = local_config.get('@password', local_config.get('password'))
    path = local_config.get('@path', local_config.get('path')) or ''
    camera_id = local_config.get(
        '@remote_camera_id', local_config.get('remote_camera_id')
    )

    logging.debug(
        'setting config for remote camera {id} on {url}'.format(
            id=camera_id, url=pretty_camera_url(local_config)
        )
    )

    ui_config = json.dumps(ui_config)

    p = path + f'/config/{camera_id}/set/'
    request = _make_request(
        scheme,
        host,
        port,
        username,
        password,
        p,
        method='POST',
        data=ui_config.encode(),
        content_type='application/json',
    )
    response = await _send_request(request)

    result = None

    if response.error:
        logging.error(
            'failed to set config for remote camera {id} on {url}: {msg}'.format(
                id=camera_id,
                url=pretty_camera_url(local_config),
                msg=utils.pretty_http_error(response),
            )
        )

        result = utils.pretty_http_error(response)

    return result


async def test(local_config, data) -> utils.CommonExternalResponse:
    scheme, host, port, username, password, path, camera_id = _remote_params(
        local_config
    )
    what = data['what']
    logging.debug(
        'testing {what} on remote camera {id}, on {url}'.format(
            what=what, id=camera_id, url=pretty_camera_url(local_config)
        )
    )

    data = json.dumps(data)

    p = path + f'/config/{camera_id}/test/'
    request = _make_request(
        scheme,
        host,
        port,
        username,
        password,
        p,
        method='POST',
        data=data,
        content_type='application/json',
    )
    response = await _send_request(request)
    if response.error:
        logging.error(
            'failed to test {what} on remote camera {id}, on {url}: {msg}'.format(
                what=what,
                id=camera_id,
                url=pretty_camera_url(local_config),
                msg=utils.pretty_http_error(response),
            )
        )

        return utils.CommonExternalResponse(
            None, error=utils.pretty_http_error(response)
        )

    return utils.CommonExternalResponse(
        None, None
    )  # it will never return result = True, what the point?


async def get_current_picture(
    local_config, width, height
) -> utils.GetCurrentPictureResponse:
    scheme, host, port, username, password, path, camera_id = _remote_params(
        local_config
    )

    logging.debug(
        'getting current picture for remote camera {id} on {url}'.format(
            id=camera_id, url=pretty_camera_url(local_config)
        )
    )

    query = {}

    if width:
        query['width'] = str(width)

    if height:
        query['height'] = str(height)

    p = path + f'/picture/{camera_id}/current/'

    request = _make_request(scheme, host, port, username, password, p, query=query)
    response = await _send_request(request)

    cookies = utils.parse_cookies(response.headers.get_list('Set-Cookie'))
    motion_detected = cookies.get('motion_detected_' + str(camera_id)) == 'true'
    capture_fps = cookies.get('capture_fps_' + str(camera_id))
    capture_fps = float(capture_fps) if capture_fps else 0
    monitor_info = cookies.get('monitor_info_' + str(camera_id))

    if response.error:
        logging.error(
            'failed to get current picture for remote camera {id} on {url}: {msg}'.format(
                id=camera_id,
                url=pretty_camera_url(local_config),
                msg=utils.pretty_http_error(response),
            )
        )

        return utils.GetCurrentPictureResponse(error=utils.pretty_http_error(response))

    return utils.GetCurrentPictureResponse(
        motion_detected=motion_detected,
        capture_fps=capture_fps,
        monitor_info=monitor_info,
        picture=response.body,
    )


async def list_media(local_config, media_type, prefix) -> utils.ListMediaResponse:
    scheme, host, port, username, password, path, camera_id = _remote_params(
        local_config
    )

    logging.debug(
        'getting media list for remote camera {id} on {url}'.format(
            id=camera_id, url=pretty_camera_url(local_config)
        )
    )

    query = {}
    if prefix is not None:
        query['prefix'] = prefix

    # timeout here is 10 times larger than usual - we expect a big delay when fetching the media list
    p = path + f'/{media_type}/{camera_id}/list/'
    request = _make_request(
        scheme,
        host,
        port,
        username,
        password,
        p,
        query=query,
        timeout=10 * settings.REMOTE_REQUEST_TIMEOUT,
    )
    response = await _send_request(request)
    if response.error:
        logging.error(
            'failed to get media list for remote camera {id} on {url}: {msg}'.format(
                id=camera_id,
                url=pretty_camera_url(local_config),
                msg=utils.pretty_http_error(response),
            )
        )

        return utils.ListMediaResponse(error=utils.pretty_http_error(response))

    try:
        response = json.loads(response.body)

    except Exception as e:
        logging.error(
            'failed to decode json answer from {url}: {msg}'.format(
                url=pretty_camera_url(local_config), msg=str(e)
            )
        )

        return utils.ListMediaResponse(error=str(e))

    return utils.ListMediaResponse(media_list=response)


async def get_media_content(
    local_config, filename, media_type
) -> utils.CommonExternalResponse:
    scheme, host, port, username, password, path, camera_id = _remote_params(
        local_config
    )

    logging.debug(
        'downloading file {filename} of remote camera {id} on {url}'.format(
            filename=filename, id=camera_id, url=pretty_camera_url(local_config)
        )
    )

    path += '/{media_type}/{id}/download/{filename}'.format(
        media_type=media_type, id=camera_id, filename=filename
    )

    # timeout here is 10 times larger than usual - we expect a big delay when fetching the media list
    request = _make_request(
        scheme,
        host,
        port,
        username,
        password,
        path,
        timeout=10 * settings.REMOTE_REQUEST_TIMEOUT,
    )
    response = await _send_request(request)
    if response.error:
        logging.error(
            'failed to download file {filename} of remote camera {id} on {url}: {msg}'.format(
                filename=filename,
                id=camera_id,
                url=pretty_camera_url(local_config),
                msg=utils.pretty_http_error(response),
            )
        )

        return utils.CommonExternalResponse(error=utils.pretty_http_error(response))

    return utils.CommonExternalResponse(result=response.body)


async def make_zipped_content(
    local_config, media_type, group
) -> utils.CommonExternalResponse:
    scheme, host, port, username, password, path, camera_id = _remote_params(
        local_config
    )

    logging.debug(
        'preparing zip file for group "{group}" of remote camera {id} on {url}'.format(
            group=group or 'ungrouped',
            id=camera_id,
            url=pretty_camera_url(local_config),
        )
    )

    prepare_path = path + '/{media_type}/{id}/zipped/{group}/'.format(
        media_type=media_type, id=camera_id, group=group
    )

    # timeout here is 100 times larger than usual - we expect a big delay
    request = _make_request(
        scheme,
        host,
        port,
        username,
        password,
        prepare_path,
        timeout=100 * settings.REMOTE_REQUEST_TIMEOUT,
    )
    response = await _send_request(request)
    if response.error:
        msg = (
            'failed to prepare zip file for group "%(group)s" '
            'of remote camera %(id)s on %(url)s: %(msg)s'
            % {
                'group': group or 'ungrouped',
                'id': camera_id,
                'url': pretty_camera_url(local_config),
                'msg': utils.pretty_http_error(response),
            }
        )

        logging.error(msg)

        return utils.CommonExternalResponse(error=utils.pretty_http_error(response))

    try:
        key = json.loads(response.body)['key']

    except Exception as e:
        logging.error(
            'failed to decode json answer from {url}: {msg}'.format(
                url=pretty_camera_url(local_config), msg=str(e)
            )
        )

        return utils.CommonExternalResponse(error=str(e))
    else:
        return utils.CommonExternalResponse(result={'key': key})


async def get_zipped_content(
    local_config, media_type, key, group
) -> utils.CommonExternalResponse:
    scheme, host, port, username, password, path, camera_id = _remote_params(
        local_config
    )

    logging.debug(
        'downloading zip file for remote camera {id} on {url}'.format(
            id=camera_id, url=pretty_camera_url(local_config)
        )
    )

    p = path + '/{media_type}/{id}/zipped/{group}/?key={key}'.format(
        media_type=media_type, group=group, id=camera_id, key=key
    )

    request = _make_request(
        scheme,
        host,
        port,
        username,
        password,
        p,
        timeout=10 * settings.REMOTE_REQUEST_TIMEOUT,
    )
    response = await _send_request(request)
    if response.error:
        logging.error(
            'failed to download zip file for remote camera {id} on {url}: {msg}'.format(
                id=camera_id,
                url=pretty_camera_url(local_config),
                msg=utils.pretty_http_error(response),
            )
        )

        return utils.CommonExternalResponse(error=utils.pretty_http_error(response))

    return utils.CommonExternalResponse(
        result={
            'data': response.body,
            'content_type': response.headers.get('Content-Type'),
            'content_disposition': response.headers.get('Content-Disposition'),
        }
    )


async def make_timelapse_movie(
    local_config, framerate, interval, group
) -> utils.CommonExternalResponse:
    scheme, host, port, username, password, path, camera_id = _remote_params(
        local_config
    )

    msg = (
        'making timelapse movie for group "%(group)s" of remote camera %(id)s '
        'with rate %(framerate)s/%(int)s on %(url)s'
        % {
            'group': group or 'ungrouped',
            'id': camera_id,
            'framerate': framerate,
            'int': interval,
            'url': pretty_camera_url(local_config),
        }
    )

    logging.debug(msg)

    path += (
        '/picture/{id}/timelapse/{group}/?interval={int}&framerate={framerate}'.format(
            id=camera_id, int=interval, framerate=framerate, group=group
        )
    )

    request = _make_request(
        scheme,
        host,
        port,
        username,
        password,
        path,
        timeout=100 * settings.REMOTE_REQUEST_TIMEOUT,
    )
    response = await _send_request(request)
    if response.error:
        msg = (
            'failed to make timelapse movie for group "%(group)s" of remote camera %(id)s '
            'with rate %(framerate)s/%(int)s on %(url)s: %(msg)s'
            % {
                'group': group or 'ungrouped',
                'id': camera_id,
                'url': pretty_camera_url(local_config),
                'int': interval,
                'framerate': framerate,
                'msg': utils.pretty_http_error(response),
            }
        )

        logging.error(msg)

        return utils.CommonExternalResponse(error=utils.pretty_http_error(response))

    try:
        response = json.loads(response.body)

    except Exception as e:
        logging.error(
            'failed to decode json answer from {url}: {msg}'.format(
                url=pretty_camera_url(local_config), msg=str(e)
            )
        )

        return utils.CommonExternalResponse(error=str(e))

    else:
        return utils.CommonExternalResponse(result=response)


async def check_timelapse_movie(local_config, group) -> utils.CommonExternalResponse:
    scheme, host, port, username, password, path, camera_id = _remote_params(
        local_config
    )

    logging.debug(
        'checking timelapse movie status for remote camera {id} on {url}'.format(
            id=camera_id, url=pretty_camera_url(local_config)
        )
    )

    p = path + '/picture/{id}/timelapse/{group}/?check=true'.format(
        id=camera_id, group=group
    )
    request = _make_request(scheme, host, port, username, password, p)
    response = await _send_request(request)

    if response.error:
        logging.error(
            'failed to check timelapse movie status for remote camera {id} on {url}: {msg}'.format(
                id=camera_id,
                url=pretty_camera_url(local_config),
                msg=utils.pretty_http_error(response),
            )
        )

        return utils.CommonExternalResponse(error=utils.pretty_http_error(response))

    try:
        response = json.loads(response.body)

    except Exception as e:
        logging.error(
            'failed to decode json answer from {url}: {msg}'.format(
                url=pretty_camera_url(local_config), msg=str(e)
            )
        )

        return utils.CommonExternalResponse(error=str(e))

    else:
        return utils.CommonExternalResponse(result=response)


async def get_timelapse_movie(local_config, key, group) -> utils.CommonExternalResponse:
    scheme, host, port, username, password, path, camera_id = _remote_params(
        local_config
    )

    logging.debug(
        'downloading timelapse movie for remote camera {id} on {url}'.format(
            id=camera_id, url=pretty_camera_url(local_config)
        )
    )

    p = path + '/picture/{id}/timelapse/{group}/?key={key}'.format(
        id=camera_id, group=group, key=key
    )

    request = _make_request(
        scheme,
        host,
        port,
        username,
        password,
        p,
        timeout=10 * settings.REMOTE_REQUEST_TIMEOUT,
    )
    response = await _send_request(request)
    if response.error:
        logging.error(
            'failed to download timelapse movie for remote camera {id} on {url}: {msg}'.format(
                id=camera_id,
                url=pretty_camera_url(local_config),
                msg=utils.pretty_http_error(response),
            )
        )

        return utils.CommonExternalResponse(error=utils.pretty_http_error(response))

    return utils.CommonExternalResponse(
        result={
            'data': response.body,
            'content_type': response.headers.get('Content-Type'),
            'content_disposition': response.headers.get('Content-Disposition'),
        }
    )


async def get_media_preview(
    local_config, filename, media_type, width, height
) -> utils.CommonExternalResponse:
    scheme, host, port, username, password, path, camera_id = _remote_params(
        local_config
    )

    logging.debug(
        'getting file preview for {filename} of remote camera {id} on {url}'.format(
            filename=filename, id=camera_id, url=pretty_camera_url(local_config)
        )
    )

    path += '/{media_type}/{id}/preview/{filename}'.format(
        media_type=media_type, id=camera_id, filename=filename
    )

    query = {}

    if width:
        query['width'] = str(width)

    if height:
        query['height'] = str(height)

    request = _make_request(scheme, host, port, username, password, path, query=query)
    response = await _send_request(request)
    if response.error:
        logging.error(
            'failed to get file preview for {filename} of remote camera {id} on {url}: {msg}'.format(
                filename=filename,
                id=camera_id,
                url=pretty_camera_url(local_config),
                msg=utils.pretty_http_error(response),
            )
        )

        return utils.CommonExternalResponse(error=utils.pretty_http_error(response))

    return utils.CommonExternalResponse(result=response.body)


async def del_media_content(
    local_config, filename, media_type
) -> utils.CommonExternalResponse:
    scheme, host, port, username, password, path, camera_id = _remote_params(
        local_config
    )

    logging.debug(
        'deleting file {filename} of remote camera {id} on {url}'.format(
            filename=filename, id=camera_id, url=pretty_camera_url(local_config)
        )
    )

    path += '/{media_type}/{id}/delete/{filename}'.format(
        media_type=media_type, id=camera_id, filename=filename
    )

    request = _make_request(
        scheme,
        host,
        port,
        username,
        password,
        path,
        method='POST',
        data='{}',
        timeout=settings.REMOTE_REQUEST_TIMEOUT,
        content_type='application/json',
    )
    response = await _send_request(request)
    if response.error:
        logging.error(
            'failed to delete file {filename} of remote camera {id} on {url}: {msg}'.format(
                filename=filename,
                id=camera_id,
                url=pretty_camera_url(local_config),
                msg=utils.pretty_http_error(response),
            )
        )

        return utils.CommonExternalResponse(error=utils.pretty_http_error(response))

    return utils.CommonExternalResponse()


async def del_media_group(
    local_config, group, media_type
) -> utils.CommonExternalResponse:
    scheme, host, port, username, password, path, camera_id = _remote_params(
        local_config
    )

    logging.debug(
        'deleting group "{group}" of remote camera {id} on {url}'.format(
            group=group or 'ungrouped',
            id=camera_id,
            url=pretty_camera_url(local_config),
        )
    )

    path += '/{media_type}/{id}/delete_all/{group}/'.format(
        media_type=media_type, id=camera_id, group=group
    )

    request = _make_request(
        scheme,
        host,
        port,
        username,
        password,
        path,
        method='POST',
        data='{}',
        timeout=settings.REMOTE_REQUEST_TIMEOUT,
        content_type='application/json',
    )
    response = await _send_request(request)
    if response.error:
        logging.error(
            'failed to delete group "{group}" of remote camera {id} on {url}: {msg}'.format(
                group=group or 'ungrouped',
                id=camera_id,
                url=pretty_camera_url(local_config),
                msg=utils.pretty_http_error(response),
            )
        )

        return utils.CommonExternalResponse(error=utils.pretty_http_error(response))

    return utils.CommonExternalResponse()


async def exec_action(local_config, action) -> utils.CommonExternalResponse:
    scheme, host, port, username, password, path, camera_id = _remote_params(
        local_config
    )

    logging.debug(
        'executing action "{action}" of remote camera {id} on {url}'.format(
            action=action, id=camera_id, url=pretty_camera_url(local_config)
        )
    )

    path += f'/action/{camera_id}/{action}/'

    request = _make_request(
        scheme,
        host,
        port,
        username,
        password,
        path,
        method='POST',
        data='{}',
        timeout=settings.REMOTE_REQUEST_TIMEOUT,
        content_type='application/json',
    )
    response = await _send_request(request)
    if response.error:
        logging.error(
            'failed to execute action "{action}" of remote camera {id} on {url}: {msg}'.format(
                action=action,
                id=camera_id,
                url=pretty_camera_url(local_config),
                msg=utils.pretty_http_error(response),
            )
        )

        return utils.CommonExternalResponse(error=utils.pretty_http_error(response))

    return utils.CommonExternalResponse()



# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from logging import debug
from subprocess import CalledProcessError

from motioneye import utils


def list_devices():
    # currently MMAL support is designed specifically for the RPi;
    # therefore we can rely on the vcgencmd to report MMAL cameras

    debug('detecting MMAL camera')

    try:
        binary = utils.call_subprocess(['which', 'vcgencmd'])

    except CalledProcessError:  # not found
        debug('unable to detect MMAL camera: vcgencmd has not been found')
        return []

    try:
        support = utils.call_subprocess([binary, 'get_camera'])

    except CalledProcessError:  # not found
        debug('unable to detect MMAL camera: "vcgencmd get_camera" failed')
        return []

    if support.startswith('supported=1 detected=1'):
        debug('MMAL camera detected')
        return [('vc.ril.camera', 'VideoCore Camera')]

    return []



# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gettext

from babel.support import Translations
from jinja2 import Environment, FileSystemLoader, select_autoescape

from motioneye import settings
from motioneye.utils.dtconv import (
    pretty_date,
    pretty_date_time,
    pretty_duration,
    pretty_time,
)

_jinja_env = None


def _init_jinja():
    global _jinja_env

    #            loader=FileSystemLoader(searchpath="templates" ),
    _jinja_env = Environment(
        loader=FileSystemLoader(settings.TEMPLATE_PATH),
        trim_blocks=False,
        extensions=['jinja2.ext.i18n'],
        autoescape=select_autoescape(['html', 'xml']),
    )
    _jinja_env.install_gettext_translations(settings.traduction, newstyle=True)

    # globals
    _jinja_env.globals['settings'] = settings

    # filters
    _jinja_env.filters['pretty_date_time'] = pretty_date_time
    _jinja_env.filters['pretty_date'] = pretty_date
    _jinja_env.filters['pretty_time'] = pretty_time
    _jinja_env.filters['pretty_duration'] = pretty_duration


def _reload_lang():
    global _jinja_env
    _jinja_env.install_gettext_translations(settings.traduction, newstyle=True)
    _jinja_env.globals['settings'] = settings


def add_template_path(path):
    global _jinja_env
    if _jinja_env is None:
        _init_jinja()

    _jinja_env.loader.searchpath.append(path)


def add_context(name, value):
    global _jinja_env
    if _jinja_env is None:
        _init_jinja()

    _jinja_env.globals[name] = value


def render(template_name, **context):
    global _jinja_env
    if _jinja_env is None:
        _init_jinja()

    template = _jinja_env.get_template(template_name)
    return template.render(**context)



# Copyright (c) 2020 Vlsarro
# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import logging
import os
import subprocess

from tornado.ioloop import IOLoop
from tornado.web import HTTPError

from motioneye import config, motionctl, remote, utils
from motioneye.handlers.base import BaseHandler

__all__ = ('ActionHandler',)


class ActionHandler(BaseHandler):
    async def post(self, camera_id, action):
        camera_id = int(camera_id)
        if camera_id not in config.get_camera_ids():
            raise HTTPError(404, 'no such camera')

        local_config = config.get_camera(camera_id)
        if utils.is_remote_camera(local_config):
            resp = await remote.exec_action(local_config, action)
            if resp.error:
                msg = (
                    'Failed to execute action on remote camera at {url}: {msg}.'.format(
                        url=remote.pretty_camera_url(local_config), msg=resp.error
                    )
                )

                return self.finish_json({'error': msg})

            return self.finish_json()

        if action == 'snapshot':
            logging.debug('executing snapshot action for camera with id %s' % camera_id)
            await self.snapshot(camera_id)
            return

        elif action == 'record_start':
            logging.debug(
                'executing record_start action for camera with id %s' % camera_id
            )
            return self.record_start(camera_id)

        elif action == 'record_stop':
            logging.debug(
                'executing record_stop action for camera with id %s' % camera_id
            )
            return self.record_stop(camera_id)

        action_commands = config.get_action_commands(local_config)
        command = action_commands.get(action)
        if not command:
            raise HTTPError(400, 'unknown action')

        logging.debug(
            f'executing {action} action for camera with id {camera_id}: "{command}"'
        )
        self.run_command_bg(command)

    def run_command_bg(self, command):
        self.p = subprocess.Popen(
            command, stderr=subprocess.STDOUT, stdout=subprocess.PIPE
        )
        self.command = command

        self.io_loop = IOLoop.current()
        self.io_loop.add_timeout(
            datetime.timedelta(milliseconds=100), self.check_command
        )

    def check_command(self):
        exit_status = self.p.poll()
        if exit_status is not None:
            output = self.p.stdout.read()
            lines = output.decode('utf-8').split('\n')
            if not lines[-1]:
                lines = lines[:-1]
            command = os.path.basename(self.command)
            if exit_status:
                logging.warning(
                    f'{command}: command has finished with non-zero exit status: {exit_status}'
                )
                for line in lines:
                    logging.warning(f'{command}: {line}')

            else:
                logging.debug('%s: command has finished' % command)
                for line in lines:
                    logging.debug(f'{command}: {line}')

            return self.finish_json({'status': exit_status})

        else:
            self.io_loop.add_timeout(
                datetime.timedelta(milliseconds=100), self.check_command
            )

    async def snapshot(self, camera_id):
        await motionctl.take_snapshot(camera_id)
        return self.finish_json({})

    def record_start(self, camera_id):
        return self.finish_json({})

    def record_stop(self, camera_id):
        return self.finish_json({})



# Copyright (c) 2020 Vlsarro
# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import functools
import logging
import re
import socket

from tornado.concurrent import Future
from tornado.ioloop import IOLoop
from tornado.iostream import IOStream

from motioneye import settings
from motioneye.utils import GetCamerasResponse, build_basic_header, cast_future
from motioneye.utils.http import RtspUrl

__all__ = ('test_rtsp_url',)


def test_rtsp_url(data: dict) -> 'Future[GetCamerasResponse]':
    url_obj = RtspUrl.from_dict(data)
    url = str(url_obj)

    called = [False]
    send_auth = [False]
    timeout = [None]
    stream = None

    io_loop = IOLoop.current()
    future = Future()

    def connect():
        if send_auth[0]:
            logging.debug(
                'testing rtsp netcam at %s (this time with credentials)' % url
            )

        else:
            logging.debug('testing rtsp netcam at %s' % url)

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        s.settimeout(settings.MJPG_CLIENT_TIMEOUT)
        stream = IOStream(s)
        stream.set_close_callback(on_close)
        f = stream.connect((url_obj.host, int(url_obj.port)))
        f.add_done_callback(on_connect)

        timeout[0] = io_loop.add_timeout(
            datetime.timedelta(seconds=settings.MJPG_CLIENT_TIMEOUT),
            functools.partial(on_connect, _timeout=True),
        )

        return stream

    def on_connect(f: Future, _timeout: bool = False) -> None:
        try:
            io_loop.remove_timeout(timeout[0])
            s = f.result()
        except Exception as e:
            logging.error(
                f'[ON_CONNECT] Stream connection error occurred: {e}', exc_info=True
            )
            if _timeout:
                return handle_error('timeout connecting to rtsp netcam')
            else:
                return handle_error('failed to connect to rtsp netcam')
        else:
            logging.debug('connected to rtsp netcam')

            lines = ['OPTIONS %s RTSP/1.0' % url, 'CSeq: 1', 'User-Agent: motionEye']

            if url_obj.username and send_auth[0]:
                auth_header = 'Authorization: ' + build_basic_header(
                    url_obj.username, url_obj.password
                )
                lines.append(auth_header)

            lines += ['', '']

            write_future = s.write('\r\n'.join(lines).encode('utf-8'))
            write_future.add_done_callback(seek_rtsp)

    def seek_rtsp(f: Future):
        try:
            f.result()
        except Exception as e:
            logging.error(f'[SEEK_RTSP] Error occurred: {e}', exc_info=True)
            if check_error():
                return
        else:
            r_future = cast_future(stream.read_until_regex(br'RTSP/1.0 \d+ '))
            r_future.add_done_callback(on_rtsp)
            timeout[0] = io_loop.add_timeout(
                datetime.timedelta(seconds=settings.MJPG_CLIENT_TIMEOUT),
                functools.partial(on_rtsp, r_future),
            )

    def on_rtsp(f: Future):
        try:
            io_loop.remove_timeout(timeout[0])
            data = f.result()
        except Exception as e:
            logging.error(f'[ON_RTSP] Error occurred: {e}', exc_info=True)
            handle_error('timeout waiting for rtsp netcam response')
        else:
            if not data:
                handle_error('timeout waiting for rtsp netcam response')

            if data.endswith(b'200 '):
                seek_server()

            elif data.endswith(b'401 '):
                if not url_obj.username or send_auth[0]:
                    # either credentials not supplied, or already sent
                    handle_error('authentication failed')

                else:
                    seek_www_authenticate()

            else:
                handle_error('rtsp netcam returned erroneous response: %s' % data)

    def seek_server():
        if check_error():
            return

        future = stream.read_until_regex(b'Server: .*')

        def f(f):
            on_server(f.result())

        future.add_done_callback(f)

        timeout[0] = io_loop.add_timeout(datetime.timedelta(seconds=1), on_server)

    def on_server(data=None):
        io_loop.remove_timeout(timeout[0])
        if data:
            identifier = re.findall('Server: (.*)', data.decode())[0].strip()
            logging.debug('rtsp netcam identifier is "%s"' % identifier)

        else:
            identifier = None
            logging.debug('no rtsp netcam identifier')

        handle_success(identifier)

    def seek_www_authenticate():
        if check_error():
            return

        f = cast_future(stream.read_until_regex(b'WWW-Authenticate: .*'))
        f.add_done_callback(on_www_authenticate)

        timeout[0] = io_loop.add_timeout(
            datetime.timedelta(seconds=1), on_www_authenticate
        )

    def on_www_authenticate(f: Future):
        auth_timeout_msg = 'timeout waiting for rtsp netcam response'
        auth_timeout_log_msg = 'timeout waiting for rtsp auth scheme'
        try:
            io_loop.remove_timeout(timeout[0])
            data = f.result()
        except Exception as e:
            logging.error(f'[ON_WWW_AUTH] {auth_timeout_log_msg}: {e}', exc_info=True)
            handle_error(f'{auth_timeout_msg}')
        else:
            if data:
                scheme = re.findall(br'WWW-Authenticate: ([^\s]+)', data)[0].strip()
                logging.debug('rtsp netcam auth scheme: %s' % scheme)
                if scheme.lower() == 'basic':
                    send_auth[0] = True
                    connect()

                else:
                    logging.debug(
                        'rtsp auth scheme digest not supported, considering credentials ok'
                    )
                    handle_success('(unknown) ')

            else:
                logging.error(f'{auth_timeout_log_msg}')
                handle_error(f'{auth_timeout_msg}')

    def on_close():
        if called[0]:
            return

        if not check_error():
            handle_error('connection closed')

    def handle_success(identifier):
        if called[0]:
            return

        called[0] = True
        cameras = []
        if identifier:
            identifier = ' ' + identifier

        else:
            identifier = ''

        cameras.append({'id': 'tcp', 'name': '%sRTSP/TCP Camera' % identifier})
        cameras.append({'id': 'udp', 'name': '%sRTSP/UDP Camera' % identifier})

        future.set_result(GetCamerasResponse(cameras, None))

    def handle_error(e):
        if called[0]:
            return

        called[0] = True
        logging.error('rtsp client error: %s' % str(e))

        try:
            stream.close()

        except Exception:
            pass

        future.set_result(GetCamerasResponse(None, error=str(e)))

    def check_error():
        error = getattr(stream, 'error', None)
        if error and getattr(error, 'strerror', None):
            handle_error(error.strerror)
            return True

        if stream and stream.socket is None:
            handle_error('connection closed')
            stream.close()

            return True

        return False

    stream = connect()
    return future



# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import fcntl
import logging
import os.path
import re
import stat
import subprocess
import time

from motioneye import utils

_resolutions_cache = {}
_ctrls_cache = {}
_ctrl_values_cache = {}

_DEV_V4L_BY_ID = '/dev/v4l/by-id/'
_V4L2_TIMEOUT = 10


def find_v4l2_ctl():
    try:
        return utils.call_subprocess(['which', 'v4l2-ctl'])

    except subprocess.CalledProcessError:  # not found
        return None


def list_devices():
    global _resolutions_cache, _ctrls_cache, _ctrl_values_cache

    logging.debug('listing V4L2 devices')
    output = b''

    try:
        output = utils.call_subprocess(
            ['v4l2-ctl', '--list-devices'], stderr=subprocess.STDOUT
        )

    except:
        logging.debug(f'v4l2-ctl error: {output}')

    name = None
    devices = []
    output = utils.make_str(output)
    for line in output.split('\n'):
        if line.startswith('\t'):
            device = line.strip()
            persistent_device = find_persistent_device(device)
            devices.append((device, persistent_device, name))

            logging.debug(f'found device {name}: {device}, {persistent_device}')

        else:
            name = line.split('(')[0].strip()

    # clear the cache
    _resolutions_cache = {}
    _ctrls_cache = {}
    _ctrl_values_cache = {}

    return devices


def list_resolutions(device):
    from motioneye import motionctl

    global _resolutions_cache

    device = utils.make_str(device)

    if device in _resolutions_cache:
        return _resolutions_cache[device]

    logging.debug(f'listing resolutions of device {device}...')

    resolutions = set()
    output = b''
    started = time.time()
    cmd = f"v4l2-ctl -d '{device}' --list-formats-ext | grep -vi stepwise | grep -oE '[0-9]+x[0-9]+' || true"
    logging.debug(f'running command "{cmd}"')

    try:
        output = utils.call_subprocess(cmd, shell=True, stderr=utils.DEV_NULL)
    except:
        logging.error(f'failed to list resolutions of device "{device}"')

    output = utils.make_str(output)

    for pair in output.split('\n'):
        pair = pair.strip()
        if not pair:
            continue

        width, height = pair.split('x')
        width = int(width)
        height = int(height)

        if (width, height) in resolutions:
            continue  # duplicate resolution

        if width < 96 or height < 96:  # some reasonable minimal values
            continue

        if not motionctl.resolution_is_valid(width, height):
            continue

        resolutions.add((width, height))

        logging.debug(f'found resolution {width}x{height} for device {device}')

    if not resolutions:
        logging.debug(f'no resolutions found for device {device}, using common values')

        # no resolution returned by v4l2-ctl call, add common default resolutions
        resolutions = utils.COMMON_RESOLUTIONS
        resolutions = [r for r in resolutions if motionctl.resolution_is_valid(*r)]

    resolutions = list(sorted(resolutions, key=lambda r: (r[0], r[1])))
    _resolutions_cache[device] = resolutions

    return resolutions


def device_present(device):
    device = utils.make_str(device)

    try:
        st = os.stat(device)
        return stat.S_ISCHR(st.st_mode)

    except:
        return False


def find_persistent_device(device):
    device = utils.make_str(device)

    try:
        devs_by_id = os.listdir(_DEV_V4L_BY_ID)

    except OSError:
        return device

    for p in devs_by_id:
        p = os.path.join(_DEV_V4L_BY_ID, p)
        if os.path.realpath(p) == device:
            return p

    return device


def list_ctrls(device):
    global _ctrls_cache

    device = utils.make_str(device)

    if device in _ctrls_cache:
        return _ctrls_cache[device]

    output = b''
    started = time.time()
    cmd = ['v4l2-ctl', '-d', device, '--list-ctrls']
    logging.debug(f'running command "{" ".join(cmd)}"')

    try:
        output = utils.call_subprocess(cmd, stderr=subprocess.STDOUT)
    except:
        logging.error(f'failed to list controls of device "{device}"')

    controls = {}
    logging.debug(f'command output: "{output}"')
    output = utils.make_str(output)
    for line in output.split('\n'):
        if not line:
            continue

        match = re.match(r'^\s*(\w+)\s+([a-f0-9x\s]+)?\(\w+\)\s*:\s*(.+)\s*', line)
        if not match:
            continue

        (control, _, properties) = match.groups()
        properties = dict(
            [v.split('=', 1) for v in properties.split(' ') if v.count('=')]
        )
        controls[control] = properties

    _ctrls_cache[device] = controls

    return controls



# Copyright (c) 2020 Vlsarro
# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import hashlib
import json
import logging

from tornado.web import HTTPError, RequestHandler

from motioneye import config, prefs, settings, template, utils

__all__ = ('BaseHandler', 'NotFoundHandler', 'ManifestHandler')


class BaseHandler(RequestHandler):
    def get_all_arguments(self) -> dict:
        keys = list(self.request.arguments.keys())
        arguments = {key: self.get_argument(key) for key in keys}

        for key in self.request.files:
            files = self.request.files[key]
            if len(files) > 1:
                arguments[key] = files

            elif len(files) > 0:
                arguments[key] = files[0]

            else:
                continue

        # consider the json passed in body as well
        data = self.get_json()
        if data and isinstance(data, dict):
            arguments.update(data)

        return arguments

    def get_json(self):
        if not hasattr(self, '_json'):
            self._json = None
            if self.request.headers.get('Content-Type', '').startswith(
                'application/json'
            ):
                self._json = json.loads(self.request.body)

        return self._json

    def get_argument(self, name, default=None, strip=True):
        def_ = {}
        argument = RequestHandler.get_argument(self, name, default=def_)
        if argument is def_:
            # try to find it in json body
            data = self.get_json()
            if data:
                argument = data.get(name, def_)

            if argument is def_:
                argument = default

        return argument

    def finish(self, chunk=None):
        if not self._finished:
            import motioneye

            self.set_header('Server', f'motionEye/{motioneye.VERSION}')

            return super().finish(chunk=chunk)
        else:
            logging.debug('Already finished')

    def render(self, template_name, content_type='text/html', **context):
        import motioneye

        self.set_header('Content-Type', content_type)

        context.setdefault('version', motioneye.VERSION)

        content = template.render(template_name, **context)
        self.finish(content)

    def finish_json(self, data=None):
        if data is None:
            data = {}
        self.set_header('Content-Type', 'application/json')
        return self.finish(json.dumps(data))

    def get_current_user(self):
        main_config = config.get_main()

        username = self.get_argument('_username', None)
        signature = self.get_argument('_signature', None)
        login = self.get_argument('_login', None) == 'true'

        admin_username = main_config.get('@admin_username')
        normal_username = main_config.get('@normal_username')

        admin_password = main_config.get('@admin_password')
        normal_password = main_config.get('@normal_password')

        admin_hash = hashlib.sha1(
            main_config['@admin_password'].encode('utf-8')
        ).hexdigest()
        normal_hash = hashlib.sha1(
            main_config['@normal_password'].encode('utf-8')
        ).hexdigest()

        if settings.HTTP_BASIC_AUTH and 'Authorization' in self.request.headers:
            up = utils.parse_basic_header(self.request.headers['Authorization'])
            if up:
                if up['username'] == admin_username and admin_password in (
                    up['password'],
                    hashlib.sha1(up['password'].encode('utf-8')).hexdigest(),
                ):
                    return 'admin'

                if up['username'] == normal_username and normal_password in (
                    up['password'],
                    hashlib.sha1(up['password'].encode('utf-8')).hexdigest(),
                ):
                    return 'normal'

        if username == admin_username and (
            signature
            == utils.compute_signature(
                self.request.method, self.request.uri, self.request.body, admin_password
            )
            or signature
            == utils.compute_signature(
                self.request.method, self.request.uri, self.request.body, admin_hash
            )
        ):
            return 'admin'

        # no authentication required for normal user
        if not username and not normal_password:
            return 'normal'

        if username == normal_username and (
            signature
            == utils.compute_signature(
                self.request.method,
                self.request.uri,
                self.request.body,
                normal_password,
            )
            or signature
            == utils.compute_signature(
                self.request.method, self.request.uri, self.request.body, normal_hash
            )
        ):
            return 'normal'

        if username and username != '_' and login:
            logging.error(f'authentication failed for user {username}')

        return None

    def get_pref(self, key):
        return prefs.get(self.current_user or 'anonymous', key)

    def set_pref(self, key, value):
        return prefs.set(self.current_user or 'anonymous', key, value)

    def _handle_request_exception(self, exception):
        try:
            if isinstance(exception, HTTPError):
                logging.error(str(exception))
                self.set_status(exception.status_code)
                self.finish_json(
                    {
                        'error': exception.log_message
                        or getattr(exception, 'reason', None)
                        or str(exception)
                    }
                )

            else:
                logging.error(str(exception), exc_info=True)
                self.set_status(500)
                self.finish_json({'error': 'internal server error'})

        except RuntimeError:
            pass  # nevermind

    @staticmethod
    def auth(admin=False, prompt=True):
        def decorator(func):
            def wrapper(self, *args, **kwargs):
                _admin = self.get_argument('_admin', None) == 'true'

                user = self.current_user
                if (user is None) or (user != 'admin' and (admin or _admin)):
                    self.set_header('Content-Type', 'application/json')
                    self.set_status(403)

                    return self.finish_json({'error': 'unauthorized', 'prompt': prompt})

                return func(self, *args, **kwargs)

            return wrapper

        return decorator

    def get(self, *args, **kwargs):
        raise HTTPError(400, 'method not allowed')

    def post(self, *args, **kwargs):
        raise HTTPError(400, 'method not allowed')

    def head(self, *args, **kwargs):
        self.finish()


class NotFoundHandler(BaseHandler):
    def get(self, *args, **kwargs):
        raise HTTPError(404, 'not found')

    post = head = get


class ManifestHandler(BaseHandler):
    def get(self):
        self.set_header('Content-Type', 'application/manifest+json')
        self.set_header('Cache-Control', 'max-age=2592000')  # 30 days
        self.render('manifest.json')



# Copyright (c) 2020 Vlsarro
# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os

from tornado.web import HTTPError

from motioneye import config, mediafiles, remote, settings, utils
from motioneye.handlers.base import BaseHandler

__all__ = ('MovieHandler',)


class MovieHandler(BaseHandler):
    async def get(self, camera_id, op, filename=None):
        if camera_id is not None:
            camera_id = int(camera_id)
            if camera_id not in config.get_camera_ids():
                raise HTTPError(404, 'no such camera')

        if op == 'list':
            await self.list(camera_id)
            return

        elif op == 'preview':
            await self.preview(camera_id, filename)
            return

        else:
            raise HTTPError(400, 'unknown operation')

    async def post(self, camera_id, op, filename=None, group=None):
        if group == '/':  # ungrouped
            group = ''

        if camera_id is not None:
            camera_id = int(camera_id)
            if camera_id not in config.get_camera_ids():
                raise HTTPError(404, 'no such camera')

        if op == 'delete':
            await self.delete(camera_id, filename)
            return

        elif op == 'delete_all':
            await self.delete_all(camera_id, group)
            return

        else:
            raise HTTPError(400, 'unknown operation')

    @BaseHandler.auth()
    async def list(self, camera_id):
        logging.debug(f'listing movies for camera {camera_id}')

        camera_config = config.get_camera(camera_id)
        if utils.is_local_motion_camera(camera_config):
            media_list = await mediafiles.list_media(
                camera_config,
                media_type='movie',
                prefix=self.get_argument('prefix', None),
            )
            if media_list is None:
                self.finish_json({'error': 'Failed to get movies list.'})

            return self.finish_json(
                {'mediaList': media_list, 'cameraName': camera_config['camera_name']}
            )

        elif utils.is_remote_camera(camera_config):
            resp = await remote.list_media(
                camera_config,
                media_type='movie',
                prefix=self.get_argument('prefix', None),
            )
            if resp.error:
                return self.finish_json(
                    {
                        'error': 'Failed to get movie list for {url}: {msg}.'.format(
                            url=remote.pretty_camera_url(camera_config), msg=resp.error
                        )
                    }
                )

            return self.finish_json(resp.media_list)

        else:  # assuming simple mjpeg camera
            raise HTTPError(400, 'unknown operation')

    @BaseHandler.auth()
    async def preview(self, camera_id, filename):
        logging.debug(
            'previewing movie {filename} of camera {id}'.format(
                filename=filename, id=camera_id
            )
        )

        camera_config = config.get_camera(camera_id)
        if utils.is_local_motion_camera(camera_config):
            content = mediafiles.get_media_preview(
                camera_config,
                filename,
                'movie',
                width=self.get_argument('width', None),
                height=self.get_argument('height', None),
            )

            if content:
                self.set_header('Content-Type', 'image/jpeg')

            else:
                self.set_header('Content-Type', 'image/svg+xml')
                content = open(
                    os.path.join(settings.STATIC_PATH, 'img', 'no-preview.svg')
                ).read()

            return self.finish(content)

        elif utils.is_remote_camera(camera_config):
            resp = await remote.get_media_preview(
                camera_config,
                filename=filename,
                media_type='movie',
                width=self.get_argument('width', None),
                height=self.get_argument('height', None),
            )

            content = resp.result
            if content:
                self.set_header('Content-Type', 'image/jpeg')

            else:
                self.set_header('Content-Type', 'image/svg+xml')
                content = open(
                    os.path.join(settings.STATIC_PATH, 'img', 'no-preview.svg')
                ).read()

            return self.finish(content)

        else:  # assuming simple mjpeg camera
            raise HTTPError(400, 'unknown operation')

    @BaseHandler.auth(admin=True)
    async def delete(self, camera_id, filename):
        logging.debug(
            'deleting movie {filename} of camera {id}'.format(
                filename=filename, id=camera_id
            )
        )

        camera_config = config.get_camera(camera_id)
        if utils.is_local_motion_camera(camera_config):
            try:
                mediafiles.del_media_content(camera_config, filename, 'movie')
                return self.finish_json()

            except Exception as e:
                return self.finish_json({'error': str(e)})

        elif utils.is_remote_camera(camera_config):
            resp = await remote.del_media_content(
                camera_config, filename=filename, media_type='movie'
            )
            if resp.error:
                return self.finish_json(
                    {
                        'error': 'Failed to delete movie from {url}: {msg}.'.format(
                            url=remote.pretty_camera_url(camera_config), msg=resp.error
                        )
                    }
                )

            return self.finish_json()

        else:  # assuming simple mjpeg camera
            raise HTTPError(400, 'unknown operation')

    @BaseHandler.auth(admin=True)
    async def delete_all(self, camera_id, group):
        logging.debug(
            'deleting movie group "{group}" of camera {id}'.format(
                group=group or 'ungrouped', id=camera_id
            )
        )

        camera_config = config.get_camera(camera_id)
        if utils.is_local_motion_camera(camera_config):
            try:
                mediafiles.del_media_group(camera_config, group, 'movie')
                return self.finish_json()

            except Exception as e:
                return self.finish_json({'error': str(e)})

        elif utils.is_remote_camera(camera_config):
            resp = await remote.del_media_group(
                camera_config, group=group, media_type='movie'
            )
            if resp.error:
                return self.finish_json(
                    {
                        'error': 'Failed to delete movie group at {url}: {msg}.'.format(
                            url=remote.pretty_camera_url(camera_config), msg=resp.error
                        )
                    }
                )

            return self.finish_json()

        else:  # assuming simple mjpeg camera
            raise HTTPError(400, 'unknown operation')



# Copyright (c) 2020 Vlsarro
# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import re
from typing import List

from tornado.concurrent import Future
from tornado.httpclient import AsyncHTTPClient, HTTPRequest, HTTPResponse

from motioneye import settings
from motioneye.utils import GetCamerasResponse, cast_future, pretty_http_error
from motioneye.utils.http import MjpegUrl

__all__ = ('test_mjpeg_url',)


def test_mjpeg_url(
    data: dict, auth_modes: List[str], allow_jpeg: bool
) -> 'Future[GetCamerasResponse]':
    url_obj = MjpegUrl.from_dict(data)
    url = str(url_obj)

    called = [False]
    status_2xx = [False]
    http_11 = [False]

    future = Future()

    def do_request() -> 'Future[HTTPResponse]':
        if url_obj.username:
            auth = auth_modes[0]

        else:
            auth = 'no'

        logging.debug(f'testing (m)jpg netcam at {url} using {auth} authentication')

        request = HTTPRequest(
            url,
            auth_username=url_obj.username,
            auth_password=url_obj.password or '',
            auth_mode=auth_modes.pop(0),
            connect_timeout=settings.REMOTE_REQUEST_TIMEOUT,
            request_timeout=settings.REMOTE_REQUEST_TIMEOUT,
            header_callback=on_header,
            validate_cert=settings.VALIDATE_CERTS,
        )

        fetch_future = cast_future(
            AsyncHTTPClient(force_instance=True).fetch(request, raise_error=False)
        )
        fetch_future.add_done_callback(on_response)
        return fetch_future

    def on_header(header: str):
        header = header.lower()
        if header.startswith('content-type') and status_2xx[0]:
            content_type = header.split(':')[1].strip()
            called[0] = True

            if content_type in ['image/jpg', 'image/jpeg', 'image/pjpg'] and allow_jpeg:
                future.set_result(
                    GetCamerasResponse(
                        [
                            {
                                'id': 1,
                                'name': 'JPEG Network Camera',
                                'keep_alive': http_11[0],
                            }
                        ],
                        None,
                    )
                )

            elif content_type.startswith('multipart/x-mixed-replace'):
                future.set_result(
                    GetCamerasResponse(
                        [
                            {
                                'id': 1,
                                'name': 'MJPEG Network Camera',
                                'keep_alive': http_11[0],
                            }
                        ],
                        None,
                    )
                )

            else:
                future.set_result(
                    GetCamerasResponse(None, error='not a supported network camera')
                )

        else:
            # check for the status header
            m = re.match(r'^http/1.(\d) (\d+) ', header)
            if m:
                if int(m.group(2)) / 100 == 2:
                    status_2xx[0] = True

                if m.group(1) == '1':
                    http_11[0] = True

    def on_response(res_future: Future):
        response = res_future.result()

        if not called[0]:
            if response.code == 401 and auth_modes and url_obj.username:
                status_2xx[0] = False
                do_request()

            else:
                called[0] = True
                error = (
                    pretty_http_error(response)
                    if response.error
                    else 'not a supported network camera'
                )
                future.set_result(GetCamerasResponse(None, error=error))

    do_request()

    return future



# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import re

from motioneye import utils


def _list_mounts():
    logging.debug('listing mounts...')

    seen_targets = set()

    mounts = []
    with open('/proc/mounts') as f:
        for line in f:
            line = line.strip()
            if not line:
                continue
            parts = line.split()
            if len(parts) < 4:
                continue

            target = parts[0]
            mount_point = parts[1]
            fstype = parts[2]
            opts = parts[3]

            if not os.access(mount_point, os.W_OK):
                continue

            if target in seen_targets:
                continue  # probably a bind mount

            seen_targets.add(target)

            if fstype == 'fuseblk':
                fstype = 'ntfs'  # most likely

            logging.debug(f'found mount "{target}" at "{mount_point}"')

            mounts.append(
                {
                    'target': target,
                    'mount_point': mount_point,
                    'fstype': fstype,
                    'opts': opts,
                }
            )

    return mounts


def _list_disks():
    if os.path.exists('/dev/disk/by-id/'):
        return _list_disks_dev_by_id()

    else:  # fall back to fdisk -l
        return _list_disks_fdisk()


def _list_disks_dev_by_id():
    logging.debug('listing disks using /dev/disk/by-id/')

    disks_by_dev = {}
    partitions_by_dev = {}

    for entry in os.listdir('/dev/disk/by-id/'):
        parts = entry.split('-', 1)
        if len(parts) < 2:
            continue

        target = os.path.realpath(os.path.join('/dev/disk/by-id/', entry))

        bus, entry = parts
        m = re.search(r'-part(\d+)$', entry)
        if m:
            part_no = int(m.group(1))
            entry = re.sub(r'-part\d+$', '', entry)

        else:
            part_no = None

        parts = entry.split('_')
        if len(parts) < 2:
            vendor = parts[0]
            model = ''

        else:
            vendor, model = parts[:2]

        if part_no is not None:
            logging.debug(
                f'found partition "{part_no}" at "{target}" on bus "{bus}": "{vendor} {model}"'
            )

            partitions_by_dev[target] = {
                'target': target,
                'bus': bus,
                'vendor': vendor,
                'model': model,
                'part_no': part_no,
                'unmatched': True,
            }

        else:
            logging.debug(
                f'found disk at "{target}" on bus "{bus}": "{vendor} {model}"'
            )

            disks_by_dev[target] = {
                'target': target,
                'bus': bus,
                'vendor': vendor,
                'model': model,
                'partitions': [],
            }

    # group partitions by disk
    for dev, partition in list(partitions_by_dev.items()):
        for disk_dev, disk in list(disks_by_dev.items()):
            if dev.startswith(disk_dev):
                disk['partitions'].append(partition)
                partition.pop('unmatched', None)

    # add separate partitions that did not match any disk
    for partition in list(partitions_by_dev.values()):
        if partition.pop('unmatched', False):
            disks_by_dev[partition['target']] = partition
            partition['partitions'] = [dict(partition)]

    # prepare flat list of disks
    disks = list(disks_by_dev.values())
    disks.sort(key=lambda d: d['vendor'])

    for disk in disks:
        disk['partitions'].sort(key=lambda p: p['part_no'])

    return disks


def _list_disks_fdisk():
    try:
        output = utils.call_subprocess(['fdisk', '-l'], stderr=utils.DEV_NULL)

    except Exception as e:
        logging.error('failed to list disks using "fdisk -l": %s' % e, exc_info=True)

        return []

    disks = []
    disk = None

    def add_disk(d):
        logging.debug(
            'found disk at "%s" on bus "%s": "%s %s"'
            % (d['target'], d['bus'], d['vendor'], d['model'])
        )

        for part in d['partitions']:
            logging.debug(
                'found partition "%s" at "%s" on bus "%s": "%s %s"'
                % (
                    part['part_no'],
                    part['target'],
                    part['bus'],
                    part['vendor'],
                    part['model'],
                )
            )

        disks.append(d)

    for line in output.split('\n'):
        line = line.replace('*', '')
        line = re.sub(r'\s+', ' ', line.strip())
        if not line:
            continue

        if line.startswith('Disk /dev/'):
            if disk and disk['partitions']:
                add_disk(disk)

            parts = line.split()

            disk = {
                'target': parts[1].strip(':'),
                'bus': '',
                'vendor': '',
                'model': parts[2] + ' ' + parts[3].strip(','),
                'partitions': [],
            }

        elif line.startswith('/dev/') and disk:
            parts = line.split()
            part_no = re.findall(r'\d+$', parts[0])
            partition = {
                'part_no': int(part_no[0]) if part_no else None,
                'target': parts[0],
                'bus': '',
                'vendor': '',
                'model': parts[4] + ' ' + ' '.join(parts[6:]),
            }

            disk['partitions'].append(partition)

    if disk and disk['partitions']:
        add_disk(disk)

    disks.sort(key=lambda d: d['target'])

    for disk in disks:
        disk['partitions'].sort(key=lambda p: p['part_no'])

    return disks


def list_mounted_disks():
    mounted_disks = []

    try:
        disks = _list_disks()
        mounts_by_target = {m['target']: m for m in _list_mounts()}

        for disk in disks:
            for partition in disk['partitions']:
                mount = mounts_by_target.get(partition['target'])
                if mount:
                    partition.update(mount)

            # filter out unmounted partitions
            disk['partitions'] = [p for p in disk['partitions'] if p.get('mount_point')]

        # filter out unmounted disks
        mounted_disks = [d for d in disks if d['partitions']]

    except Exception as e:
        logging.error('failed to list mounted disks: %s' % e, exc_info=True)

    return mounted_disks


def list_mounted_partitions():
    mounted_partitions = {}

    try:
        disks = _list_disks()
        mounts_by_target = {m['target']: m for m in _list_mounts()}

        for disk in disks:
            for partition in disk['partitions']:
                mount = mounts_by_target.get(partition['target'])
                if mount:
                    partition.update(mount)
                    mounted_partitions[partition['target']] = partition

    except Exception as e:
        logging.error('failed to list mounted partitions: %s' % e, exc_info=True)

    return mounted_partitions



# Copyright (c) 2020 Vlsarro
# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
from typing import Union

__all__ = ('pretty_date_time', 'pretty_date', 'pretty_duration', 'pretty_time')


def pretty_date_time(date_time, tzinfo=None, short=False):
    if date_time is None:
        return '(' + _('neniam') + ')'

    if isinstance(date_time, int):
        return pretty_date_time(datetime.datetime.fromtimestamp(date_time))

    if short:
        text = '{day} {month}, {hm}'.format(
            day=date_time.day,
            month=date_time.strftime('%b'),
            hm=date_time.strftime('%H:%M'),
        )

    else:
        text = '{day} {month} {year}, {hm}'.format(
            day=date_time.day,
            month=date_time.strftime('%B'),
            year=date_time.year,
            hm=date_time.strftime('%H:%M'),
        )

    if tzinfo:
        offset = tzinfo.utcoffset(datetime.datetime.utcnow()).seconds
        tz = 'GMT'
        if offset >= 0:
            tz += '+'

        else:
            tz += '-'
            offset = -offset

        tz += '%.2d' % (offset // 3600) + ':%.2d' % ((offset % 3600) // 60)

        text += ' (' + tz + ')'

    return text


def pretty_date(d: Union[datetime.date, int]) -> str:
    if d is None:
        return '(' + _('neniam') + ')'

    if isinstance(d, int):
        return pretty_date(datetime.datetime.fromtimestamp(d))

    return '{day} {month} {year}'.format(
        day=d.day, month=_(d.strftime('%B')), year=d.year
    )


def pretty_time(t: Union[datetime.time, datetime.timedelta]) -> str:
    if t is None:
        return ''

    if isinstance(t, datetime.timedelta):
        hour = t.seconds // 3600
        minute = (t.seconds % 3600) // 60
        t = datetime.time(hour=hour, minute=minute)

    return '{hm}'.format(hm=t.strftime('%H:%M'))


def pretty_duration(duration):
    if duration is None:
        duration = 0

    if isinstance(duration, datetime.timedelta):
        duration = duration.seconds + duration.days * 86400

    if duration < 0:
        negative = True
        duration = -duration

    else:
        negative = False

    days = duration // 86400
    duration %= 86400
    hours = duration // 3600
    duration %= 3600
    minutes = duration // 60
    duration %= 60
    seconds = duration

    # treat special cases
    special_result = None
    if days != 0 and hours == 0 and minutes == 0 and seconds == 0:
        if days == 1:
            special_result = str(days) + ' ' + _('tago')

        elif days == 7:
            special_result = '1 ' + _('semajno')

        elif days in [30, 31, 32]:
            special_result = '1 ' + _('monato')

        elif days in [365, 366]:
            special_result = '1 ' + _('jaro')

        else:
            special_result = str(days) + ' ' + _('tagoj')

    elif days == 0 and hours != 0 and minutes == 0 and seconds == 0:
        if hours == 1:
            special_result = str(hours) + ' ' + _('horo')

        else:
            special_result = str(hours) + ' ' + _('horoj')

    elif days == 0 and hours == 0 and minutes != 0 and seconds == 0:
        if minutes == 1:
            special_result = str(minutes) + ' ' + _('minuto')

        else:
            special_result = str(minutes) + ' ' + _('minutoj')

    elif days == 0 and hours == 0 and minutes == 0 and seconds != 0:
        if seconds == 1:
            special_result = str(seconds) + ' ' + _('sekundo')

        else:
            special_result = str(seconds) + ' ' + _('sekundoj')

    elif days == 0 and hours == 0 and minutes == 0 and seconds == 0:
        special_result = str(0) + ' ' + _('sekundoj')

    if special_result:
        if negative:
            special_result = _('minus') + ' ' + special_result

        return special_result

    if days:
        fmt = "{d}d{h}h{m}m"

    elif hours:
        fmt = "{h}h{m}m"

    elif minutes:
        fmt = "{m}m"
        if seconds:
            fmt += "{s}s"

    else:
        fmt = "{s}s"

    if negative:
        fmt = '-' + fmt

    return fmt.format(d=days, h=hours, m=minutes, s=seconds)



# Copyright (c) 2020 Vlsarro
# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import logging
import os
import re

from tornado.ioloop import IOLoop
from tornado.web import HTTPError

from motioneye import (
    config,
    mediafiles,
    mjpgclient,
    monitor,
    motionctl,
    remote,
    settings,
    utils,
)
from motioneye.handlers.base import BaseHandler

__all__ = ('PictureHandler',)


class PictureHandler(BaseHandler):
    def compute_etag(self):
        return None

    async def get(self, camera_id, op, filename=None, group=None):
        if camera_id is not None:
            camera_id = int(camera_id)
            if camera_id not in config.get_camera_ids():
                raise HTTPError(404, 'no such camera')

        if op == 'current':
            await self.current(camera_id)

        elif op == 'list':
            await self.list(camera_id)

        elif op == 'frame':
            await self.frame(camera_id)

        elif op == 'download':
            await self.download(camera_id, filename)

        elif op == 'preview':
            await self.preview(camera_id, filename)

        elif op == 'zipped':
            await self.zipped(camera_id, group)

        elif op == 'timelapse':
            await self.timelapse(camera_id, group)

        else:
            raise HTTPError(400, 'unknown operation')

    async def post(self, camera_id, op, filename=None, group=None):
        if group == '/':  # ungrouped
            group = ''

        if camera_id is not None:
            camera_id = int(camera_id)
            if camera_id not in config.get_camera_ids():
                raise HTTPError(404, 'no such camera')

        if op == 'delete':
            await self.delete(camera_id, filename)

        elif op == 'delete_all':
            await self.delete_all(camera_id, group)

        else:
            raise HTTPError(400, 'unknown operation')

    @BaseHandler.auth(prompt=False)
    async def current(self, camera_id, retry=0):
        self.set_header('Content-Type', 'image/jpeg')
        self.set_header('Cache-Control', 'no-store, must-revalidate')
        self.set_header('Pragma', 'no-cache')
        self.set_header('Expires', '0')

        width = self.get_argument('width', None)
        height = self.get_argument('height', None)

        width = width and float(width)
        height = height and float(height)

        camera_id_str = str(camera_id)

        camera_config = config.get_camera(camera_id)
        if utils.is_local_motion_camera(camera_config):
            picture = mediafiles.get_current_picture(
                camera_config, width=width, height=height
            )

            # picture is not available usually when the corresponding internal mjpeg client has been closed;
            # get_current_picture() will make sure to start a client, but a jpeg frame is not available right away;
            # wait at most 5 seconds and retry every 200 ms.
            if not picture and retry < 25:
                return IOLoop.current().add_timeout(
                    datetime.timedelta(seconds=0.2),
                    self.current,
                    camera_id=camera_id,
                    retry=retry + 1,
                )

            self.set_cookie(
                'motion_detected_' + camera_id_str,
                str(motionctl.is_motion_detected(camera_id)).lower(),
            )
            self.set_cookie(
                'capture_fps_' + camera_id_str, '%.1f' % mjpgclient.get_fps(camera_id)
            )
            self.set_cookie(
                'monitor_info_' + camera_id_str, monitor.get_monitor_info(camera_id)
            )

            return self.try_finish(picture)

        elif utils.is_remote_camera(camera_config):
            resp = await remote.get_current_picture(
                camera_config, width=width, height=height
            )
            if resp.error:
                return self.try_finish(None)

            self.set_cookie(
                'motion_detected_' + camera_id_str, str(resp.motion_detected).lower()
            )
            self.set_cookie('capture_fps_' + camera_id_str, '%.1f' % resp.capture_fps)
            self.set_cookie('monitor_info_' + camera_id_str, resp.monitor_info or '')

            return self.try_finish(resp.picture)

        else:  # assuming simple mjpeg camera
            raise HTTPError(400, 'unknown operation')

    @BaseHandler.auth()
    async def list(self, camera_id):
        logging.debug(f'listing pictures for camera {camera_id}')

        camera_config = config.get_camera(camera_id)
        if utils.is_local_motion_camera(camera_config):
            media_list = await mediafiles.list_media(
                camera_config,
                media_type='picture',
                prefix=self.get_argument('prefix', None),
            )
            if media_list is None:
                self.finish_json({'error': 'Failed to get movies list.'})

            return self.finish_json(
                {'mediaList': media_list, 'cameraName': camera_config['camera_name']}
            )

        elif utils.is_remote_camera(camera_config):
            resp = await remote.list_media(
                camera_config,
                media_type='picture',
                prefix=self.get_argument('prefix', None),
            )
            if resp.error:
                return self.finish_json(
                    {
                        'error': 'Failed to get picture list for {url}: {msg}.'.format(
                            url=remote.pretty_camera_url(camera_config), msg=resp.error
                        )
                    }
                )

            return self.finish_json(resp.media_list)

        else:  # assuming simple mjpeg camera
            raise HTTPError(400, 'unknown operation')

    async def frame(self, camera_id):
        camera_config = config.get_camera(camera_id)

        if (
            utils.is_local_motion_camera(camera_config)
            or utils.is_simple_mjpeg_camera(camera_config)
            or self.get_argument('title', None) is not None
        ):
            return self.render(
                'main.html',
                frame=True,
                camera_id=camera_id,
                camera_config=camera_config,
                title=self.get_argument('title', camera_config.get('camera_name', '')),
                admin_username=config.get_main().get('@admin_username'),
                static_path='../../../static/',
            )

        elif utils.is_remote_camera(camera_config):
            resp = await remote.get_config(camera_config)
            if resp.error:
                return self.render(
                    'main.html',
                    frame=True,
                    camera_id=camera_id,
                    camera_config=camera_config,
                    title=self.get_argument('title', ''),
                )

            # issue a fake motion_camera_ui_to_dict() call to transform
            # the remote UI values into motion config directives
            remote_config = config.motion_camera_ui_to_dict(resp.remote_ui_config)

            return self.render(
                'main.html',
                frame=True,
                camera_id=camera_id,
                camera_config=remote_config,
                title=self.get_argument('title', remote_config['camera_name']),
                admin_username=config.get_main().get('@admin_username'),
            )

    @BaseHandler.auth()
    async def download(self, camera_id, filename):
        logging.debug(
            'downloading picture {filename} of camera {id}'.format(
                filename=filename, id=camera_id
            )
        )

        camera_config = config.get_camera(camera_id)
        if utils.is_local_motion_camera(camera_config):
            content = mediafiles.get_media_content(camera_config, filename, 'picture')

            pretty_filename = (
                camera_config['camera_name'] + '_' + os.path.basename(filename)
            )
            self.set_header('Content-Type', 'image/jpeg')
            self.set_header(
                'Content-Disposition', 'attachment; filename=' + pretty_filename + ';'
            )

            return self.finish(content)

        elif utils.is_remote_camera(camera_config):
            resp = await remote.get_media_content(
                camera_config, filename=filename, media_type='picture'
            )
            if resp.error:
                return self.finish_json(
                    {
                        'error': 'Failed to download picture from {url}: {msg}.'.format(
                            url=remote.pretty_camera_url(camera_config), msg=resp.error
                        )
                    }
                )

            pretty_filename = os.path.basename(
                filename
            )  # no camera name available w/o additional request
            self.set_header('Content-Type', 'image/jpeg')
            self.set_header(
                'Content-Disposition', 'attachment; filename=' + pretty_filename + ';'
            )

            return self.finish(resp.result)

        else:  # assuming simple mjpeg camera
            raise HTTPError(400, 'unknown operation')

    @BaseHandler.auth()
    async def preview(self, camera_id, filename):
        logging.debug(
            'previewing picture {filename} of camera {id}'.format(
                filename=filename, id=camera_id
            )
        )

        camera_config = config.get_camera(camera_id)
        if utils.is_local_motion_camera(camera_config):
            content = mediafiles.get_media_preview(
                camera_config,
                filename,
                'picture',
                width=self.get_argument('width', None),
                height=self.get_argument('height', None),
            )

            if content:
                self.set_header('Content-Type', 'image/jpeg')

            else:
                self.set_header('Content-Type', 'image/svg+xml')
                content = open(
                    os.path.join(settings.STATIC_PATH, 'img', 'no-preview.svg'), 'rb'
                ).read()

            return self.finish(content)

        elif utils.is_remote_camera(camera_config):
            resp = await remote.get_media_preview(
                camera_config,
                filename=filename,
                media_type='picture',
                width=self.get_argument('width', None),
                height=self.get_argument('height', None),
            )
            content = resp.result
            if content:
                self.set_header('Content-Type', 'image/jpeg')

            else:
                self.set_header('Content-Type', 'image/svg+xml')
                content = open(
                    os.path.join(settings.STATIC_PATH, 'img', 'no-preview.svg')
                ).read()

            return self.finish(content)

        else:  # assuming simple mjpeg camera
            raise HTTPError(400, 'unknown operation')

    @BaseHandler.auth(admin=True)
    async def delete(self, camera_id, filename):
        logging.debug(
            'deleting picture {filename} of camera {id}'.format(
                filename=filename, id=camera_id
            )
        )

        camera_config = config.get_camera(camera_id)
        if utils.is_local_motion_camera(camera_config):
            try:
                mediafiles.del_media_content(camera_config, filename, 'picture')
                return self.finish_json()

            except Exception as e:
                return self.finish_json({'error': str(e)})

        elif utils.is_remote_camera(camera_config):
            resp = await remote.del_media_content(
                camera_config, filename=filename, media_type='picture'
            )
            if resp.error:
                return self.finish_json(
                    {
                        'error': 'Failed to delete picture from {url}: {msg}.'.format(
                            url=remote.pretty_camera_url(camera_config), msg=resp.error
                        )
                    }
                )

            return self.finish_json()

        else:  # assuming simple mjpeg camera
            raise HTTPError(400, 'unknown operation')

    @BaseHandler.auth()
    async def zipped(self, camera_id, group):
        key = self.get_argument('key', None)
        camera_config = config.get_camera(camera_id)

        if key:
            logging.debug(
                'serving zip file for group "{group}" of camera {id} with key {key}'.format(
                    group=group or 'ungrouped', id=camera_id, key=key
                )
            )

            if utils.is_local_motion_camera(camera_config):
                data = mediafiles.get_prepared_cache(key)
                if not data:
                    logging.error(
                        'prepared cache data for key "%s" does not exist' % key
                    )

                    raise HTTPError(404, 'no such key')

                pretty_filename = camera_config['camera_name'] + '_' + group
                pretty_filename = re.sub('[^a-zA-Z0-9]', '_', pretty_filename)

                self.set_header('Content-Type', 'application/zip')
                self.set_header(
                    'Content-Disposition',
                    'attachment; filename=' + pretty_filename + '.zip;',
                )
                return self.finish(data)

            elif utils.is_remote_camera(camera_config):
                resp = await remote.get_zipped_content(
                    camera_config, media_type='picture', key=key, group=group
                )
                if resp.error:
                    return self.finish_json(
                        {
                            'error': 'Failed to download zip file from {url}: {msg}.'.format(
                                url=remote.pretty_camera_url(camera_config),
                                msg=resp.error,
                            )
                        }
                    )

                self.set_header('Content-Type', resp.result['content_type'])
                self.set_header(
                    'Content-Disposition', resp.result['content_disposition']
                )
                return self.finish(resp.result['data'])

            else:  # assuming simple mjpeg camera
                raise HTTPError(400, 'unknown operation')

        else:  # prepare
            logging.debug(
                'preparing zip file for group "{group}" of camera {id}'.format(
                    group=group or 'ungrouped', id=camera_id
                )
            )

            if utils.is_local_motion_camera(camera_config):
                data = await mediafiles.get_zipped_content(
                    camera_config, media_type='picture', group=group
                )
                if data is None:
                    return self.finish_json({'error': 'Failed to create zip file.'})

                key = mediafiles.set_prepared_cache(data)
                logging.debug(
                    'prepared zip file for group "{group}" of camera {id} with key {key}'.format(
                        group=group or 'ungrouped', id=camera_id, key=key
                    )
                )
                self.finish_json({'key': key})

            elif utils.is_remote_camera(camera_config):
                resp = await remote.make_zipped_content(
                    camera_config, media_type='picture', group=group
                )
                if resp.error:
                    return self.finish_json(
                        {
                            'error': 'Failed to make zip file at {url}: {msg}.'.format(
                                url=remote.pretty_camera_url(camera_config),
                                msg=resp.error,
                            )
                        }
                    )

                return self.finish_json({'key': resp.result['key']})

            else:  # assuming simple mjpeg camera
                raise HTTPError(400, 'unknown operation')

    @BaseHandler.auth()
    async def timelapse(self, camera_id, group):
        key = self.get_argument('key', None)
        check = self.get_argument('check', False)
        camera_config = config.get_camera(camera_id)

        if key:  # download
            logging.debug(
                'serving timelapse movie for group "{group}" of camera {id} with key {key}'.format(
                    group=group or 'ungrouped', id=camera_id, key=key
                )
            )

            if utils.is_local_motion_camera(camera_config):
                data = mediafiles.get_prepared_cache(key)
                if data is None:
                    logging.error(
                        'prepared cache data for key "%s" does not exist' % key
                    )

                    raise HTTPError(404, 'no such key')

                pretty_filename = camera_config['camera_name'] + '_' + group
                pretty_filename = re.sub('[^a-zA-Z0-9]', '_', pretty_filename)
                filename_ext = mediafiles.FFMPEG_EXT_MAPPING.get(
                    camera_config['movie_codec'], 'avi'
                )
                pretty_filename += '.' + filename_ext

                self.set_header(
                    'Content-Type',
                    mediafiles.MOVIE_EXT_TYPE_MAPPING.get(
                        filename_ext, 'video/x-msvideo'
                    ),
                )
                self.set_header(
                    'Content-Disposition',
                    'attachment; filename=' + pretty_filename + ';',
                )
                return self.finish(data)

            elif utils.is_remote_camera(camera_config):
                resp = await remote.get_timelapse_movie(camera_config, key, group=group)
                if resp.error:
                    msg = (
                        'Failed to download timelapse movie from {url}: {msg}.'.format(
                            url=remote.pretty_camera_url(camera_config), msg=resp.error
                        )
                    )

                    return self.finish_json({'error': msg})

                self.set_header('Content-Type', resp.result['content_type'])
                self.set_header(
                    'Content-Disposition', resp.result['content_disposition']
                )
                return self.finish(resp.result['data'])

            else:  # assuming simple mjpeg camera
                raise HTTPError(400, 'unknown operation')

        elif check:
            logging.debug(
                'checking timelapse movie status for group "{group}" of camera {id}'.format(
                    group=group or 'ungrouped', id=camera_id
                )
            )

            if utils.is_local_motion_camera(camera_config):
                status = mediafiles.check_timelapse_movie()
                if status['progress'] == -1 and status['data']:
                    key = mediafiles.set_prepared_cache(status['data'])
                    logging.debug(
                        'prepared timelapse movie for group "{group}" of camera {id} with key {key}'.format(
                            group=group or 'ungrouped', id=camera_id, key=key
                        )
                    )
                    return self.finish_json({'key': key, 'progress': -1})

                else:
                    return self.finish_json(status)

            elif utils.is_remote_camera(camera_config):
                resp = await remote.check_timelapse_movie(camera_config, group=group)
                if resp.error:
                    msg = 'Failed to check timelapse movie progress at {url}: {msg}.'.format(
                        url=remote.pretty_camera_url(camera_config), msg=resp.error
                    )

                    return self.finish_json({'error': msg})

                if resp.result['progress'] == -1 and resp.result.get('key'):
                    self.finish_json({'key': resp.result['key'], 'progress': -1})

                else:
                    self.finish_json(resp.result)

            else:  # assuming simple mjpeg camera
                raise HTTPError(400, 'unknown operation')

        else:  # start timelapse
            interval = int(self.get_argument('interval'))
            framerate = int(self.get_argument('framerate'))

            msg = 'preparing timelapse movie for group "{group}" of camera {id} with rate {framerate}/{int}'.format(
                group=group or 'ungrouped',
                id=camera_id,
                framerate=framerate,
                int=interval,
            )
            logging.debug(msg)

            if utils.is_local_motion_camera(camera_config):
                status = mediafiles.check_timelapse_movie()
                if status['progress'] != -1:
                    return self.finish_json(
                        {'progress': status['progress']}
                    )  # timelapse already active

                else:
                    mediafiles.make_timelapse_movie(
                        camera_config, framerate, interval, group=group
                    )
                    return self.finish_json({'progress': -1})

            elif utils.is_remote_camera(camera_config):
                check_timelapse_resp = await remote.check_timelapse_movie(
                    camera_config, group=group
                )
                if check_timelapse_resp.error:
                    return self.finish_json(
                        {
                            'error': 'Failed to make timelapse movie at {url}: {msg}.'.format(
                                url=remote.pretty_camera_url(camera_config),
                                msg=check_timelapse_resp.error,
                            )
                        }
                    )

                if check_timelapse_resp.result['progress'] != -1:
                    # timelapse already active
                    return self.finish_json(
                        {'progress': check_timelapse_resp.result['progress']}
                    )

                make_timelapse_resp = await remote.make_timelapse_movie(
                    camera_config, framerate, interval, group=group
                )
                if make_timelapse_resp.error:
                    return self.finish_json(
                        {
                            'error': 'Failed to make timelapse movie at {url}: {msg}.'.format(
                                url=remote.pretty_camera_url(camera_config),
                                msg=make_timelapse_resp.error,
                            )
                        }
                    )

                return self.finish_json({'progress': -1})

            else:  # assuming simple mjpeg camera
                raise HTTPError(400, 'unknown operation')

    @BaseHandler.auth(admin=True)
    async def delete_all(self, camera_id, group):
        logging.debug(
            'deleting picture group "{group}" of camera {id}'.format(
                group=group or 'ungrouped', id=camera_id
            )
        )

        camera_config = config.get_camera(camera_id)
        if utils.is_local_motion_camera(camera_config):
            try:
                mediafiles.del_media_group(camera_config, group, 'picture')
                return self.finish_json()

            except Exception as e:
                return self.finish_json({'error': str(e)})

        elif utils.is_remote_camera(camera_config):
            resp = await remote.del_media_group(
                camera_config, group=group, media_type='picture'
            )
            if resp.error:
                return self.finish_json(
                    {
                        'error': 'Failed to delete picture group at {url}: {msg}.'.format(
                            url=remote.pretty_camera_url(camera_config), msg=resp.error
                        )
                    }
                )

            return self.finish_json()

        else:  # assuming simple mjpeg camera
            raise HTTPError(400, 'unknown operation')

    def try_finish(self, content):
        try:
            return self.finish(content)

        except OSError as e:
            logging.warning(f'could not write response: {str(e)}')



# Copyright (c) 2020 Vlsarro
# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os

from tornado.web import HTTPError

from motioneye import settings, utils
from motioneye.handlers.base import BaseHandler

__all__ = ('LogHandler',)


class LogHandler(BaseHandler):
    LOGS = {
        'motion': (os.path.join(settings.LOG_PATH, 'motion.log'), 'motion.log'),
    }

    @BaseHandler.auth(admin=True)
    def get(self, name):
        log = self.LOGS.get(name)
        if log is None:
            raise HTTPError(404, 'no such log')

        (path, filename) = log

        self.set_header('Content-Type', 'text/plain')
        self.set_header('Content-Disposition', 'attachment; filename=' + filename + ';')

        if path.startswith('/'):  # an actual path
            logging.debug(f'serving log file "{filename}" from "{path}"')

            with open(path) as f:
                self.finish(f.read())

        else:  # a command to execute
            logging.debug(f'serving log file "{filename}" from command "{path}"')

            try:
                output = utils.call_subprocess(path.split())

            except Exception as e:
                output = 'failed to execute command: %s' % e

            self.finish(output)



# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import errno
import logging
import os.path
import re
import signal
import subprocess
import time

from tornado.httpclient import AsyncHTTPClient, HTTPRequest
from tornado.ioloop import IOLoop

from motioneye import mediafiles, settings, update, utils
from motioneye.controls.powerctl import PowerControl

_MOTION_CONTROL_TIMEOUT = 5

_started = False
_motion_binary_cache = None
_motion_detected = {}


def find_motion():
    global _motion_binary_cache
    if _motion_binary_cache:
        return _motion_binary_cache

    if settings.MOTION_BINARY:
        if os.path.exists(settings.MOTION_BINARY):
            binary = settings.MOTION_BINARY

        else:
            return None, None

    else:  # autodetect motion binary path
        try:
            binary = utils.call_subprocess(['which', 'motion'])

        except subprocess.CalledProcessError:  # not found
            return None, None

    try:
        help = utils.call_subprocess(binary + ' -h || true', shell=True)

    except subprocess.CalledProcessError:  # not found
        return None, None

    result = re.findall('motion Version ([^,]+)', help, re.IGNORECASE)
    version = result and result[0] or ''

    logging.debug(f'using motion version {version}')

    _motion_binary_cache = (binary, version)

    return _motion_binary_cache


def start(deferred=False):
    from motioneye import config, mjpgclient

    if deferred:
        io_loop = IOLoop.current()
        io_loop.add_callback(start, deferred=False)

    global _started

    _started = True

    enabled_local_motion_cameras = config.get_enabled_local_motion_cameras()
    if running() or not enabled_local_motion_cameras:
        return

    logging.debug('starting motion')

    program = find_motion()
    if not program[0]:
        raise Exception('motion executable could not be found')

    program, version = program  # @UnusedVariable

    logging.debug(f'starting motion binary "{program}"')

    motion_config_path = os.path.join(settings.CONF_PATH, 'motion.conf')
    motion_log_path = os.path.join(settings.LOG_PATH, 'motion.log')
    motion_pid_path = os.path.join(settings.RUN_PATH, 'motion.pid')

    args = [program, '-n', '-c', motion_config_path, '-d']

    if settings.LOG_LEVEL <= logging.DEBUG:
        args.append('9')

    elif settings.LOG_LEVEL <= logging.WARN:
        args.append('5')

    elif settings.LOG_LEVEL <= logging.ERROR:
        args.append('4')

    else:  # fatal, quiet
        args.append('1')

    log_file = open(motion_log_path, 'w')

    process = subprocess.Popen(
        args, stdout=log_file, stderr=log_file, close_fds=True, cwd=settings.CONF_PATH
    )

    # wait 2 seconds to see that the process has successfully started
    for i in range(20):  # @UnusedVariable
        time.sleep(0.1)
        exit_code = process.poll()
        if exit_code is not None and exit_code != 0:
            raise Exception('motion failed to start')

    pid = process.pid

    # write the pid to file
    with open(motion_pid_path, 'w') as f:
        f.write(str(pid) + '\n')

    IOLoop.current().spawn_callback(_disable_initial_motion_detection)

    # if mjpg client idle timeout is disabled, create mjpg clients for all cameras by default
    if not settings.MJPG_CLIENT_IDLE_TIMEOUT:
        logging.debug('creating default mjpg clients for local cameras')
        for camera in enabled_local_motion_cameras:
            mjpgclient.get_jpg(camera['@id'])


def stop(invalidate=False):
    from motioneye import mjpgclient

    global _started

    _started = False

    if not running():
        return

    logging.debug('stopping motion')

    mjpgclient.close_all(invalidate=invalidate)

    pid = _get_pid()
    if pid is not None:
        try:
            # send the TERM signal once
            os.kill(pid, signal.SIGTERM)

            # wait 5 seconds for the process to exit
            for i in range(50):  # @UnusedVariable
                os.waitpid(pid, os.WNOHANG)
                time.sleep(0.1)

            # send the KILL signal once
            os.kill(pid, signal.SIGKILL)

            # wait 2 seconds for the process to exit
            for i in range(20):  # @UnusedVariable
                time.sleep(0.1)
                os.waitpid(pid, os.WNOHANG)

            # the process still did not exit
            if settings.ENABLE_REBOOT:
                logging.error('could not terminate the motion process')
                PowerControl.reboot()

            else:
                raise Exception('could not terminate the motion process')

        except OSError as e:
            if e.errno not in (errno.ESRCH, errno.ECHILD):
                raise


def running():
    pid = _get_pid()
    if pid is None:
        return False

    try:
        os.waitpid(pid, os.WNOHANG)
        os.kill(pid, 0)

        # the process is running
        return True

    except OSError as e:
        if e.errno not in (errno.ESRCH, errno.ECHILD):
            raise

    return False


def started():
    return _started


async def get_motion_detection(camera_id) -> utils.GetMotionDetectionResult:
    motion_camera_id = camera_id_to_motion_camera_id(camera_id)
    if motion_camera_id is None:
        error = f'could not find motion camera id for camera with id {camera_id}'
        logging.error(error)
        return utils.GetMotionDetectionResult(None, error=error)

    url = f'http://127.0.0.1:{settings.MOTION_CONTROL_PORT}/{motion_camera_id}/detection/status'

    request = HTTPRequest(
        url,
        connect_timeout=_MOTION_CONTROL_TIMEOUT,
        request_timeout=_MOTION_CONTROL_TIMEOUT,
    )
    resp = await AsyncHTTPClient().fetch(request)
    if resp.error:
        return utils.GetMotionDetectionResult(None, error=utils.pretty_http_error(resp))

    resp_body = resp.body.decode('utf-8')
    enabled = bool(resp_body.count('active'))

    logging.debug(
        f"motion detection is {['disabled', 'enabled'][enabled]} for camera with id {id}"
    )

    return utils.GetMotionDetectionResult(enabled, None)


async def set_motion_detection(camera_id, enabled):
    motion_camera_id = camera_id_to_motion_camera_id(camera_id)
    if motion_camera_id is None:
        return logging.error(
            f'could not find motion camera id for camera with id {camera_id}'
        )

    if not enabled:
        _motion_detected[camera_id] = False

    logging.debug(
        f"{['disabling', 'enabling'][enabled]} motion detection for camera with id {camera_id}"
    )

    url = f"http://127.0.0.1:{settings.MOTION_CONTROL_PORT}/{motion_camera_id}/detection/{['pause', 'start'][enabled]}"

    request = HTTPRequest(
        url,
        connect_timeout=_MOTION_CONTROL_TIMEOUT,
        request_timeout=_MOTION_CONTROL_TIMEOUT,
    )
    resp = await AsyncHTTPClient().fetch(request)
    if resp.error:
        logging.error(
            'failed to {} motion detection for camera with id {}: {}'.format(
                ['disable', 'enable'][enabled],
                camera_id,
                utils.pretty_http_error(resp),
            )
        )

    else:
        logging.debug(
            f"successfully {['disabled', 'enabled'][enabled]} motion detection for camera with id {camera_id}"
        )


async def take_snapshot(camera_id):
    motion_camera_id = camera_id_to_motion_camera_id(camera_id)
    if motion_camera_id is None:
        return logging.error(
            f'could not find motion camera id for camera with id {camera_id}'
        )

    logging.debug(f'taking snapshot for camera with id {camera_id}')

    url = f'http://127.0.0.1:{settings.MOTION_CONTROL_PORT}/{motion_camera_id}/action/snapshot'

    request = HTTPRequest(
        url,
        connect_timeout=_MOTION_CONTROL_TIMEOUT,
        request_timeout=_MOTION_CONTROL_TIMEOUT,
    )
    resp = await AsyncHTTPClient().fetch(request)
    if resp.error:
        logging.error(
            f'failed to take snapshot for camera with id {camera_id}: {utils.pretty_http_error(resp)}'
        )

    else:
        logging.debug(f'successfully took snapshot for camera with id {camera_id}')


def is_motion_detected(camera_id):
    return _motion_detected.get(camera_id, False)


def set_motion_detected(camera_id, motion_detected):
    if motion_detected:
        logging.debug(f'marking motion detected for camera with id {camera_id}')

    else:
        logging.debug(f'clearing motion detected for camera with id {camera_id}')

    _motion_detected[camera_id] = motion_detected


def camera_id_to_motion_camera_id(camera_id):
    from motioneye import config

    # find the corresponding motion camera_id
    # (which can be different from camera_id)

    main_config = config.get_main()
    cameras = main_config.get('camera', [])

    camera_filename = f'camera-{camera_id}.conf'
    for i, camera in enumerate(cameras):
        if camera != camera_filename:
            continue

        return i + 1

    return None


def motion_camera_id_to_camera_id(motion_camera_id):
    from motioneye import config

    main_config = config.get_main()
    cameras = main_config.get('camera', [])

    try:
        return int(
            re.search(r'camera-(\d+).conf', cameras[int(motion_camera_id) - 1]).group(1)
        )

    except IndexError:
        return None


def is_motion_pre42():
    binary, version = find_motion()
    if not binary:
        return False

    return update.compare_versions(version, '4.2') < 0


def is_motion_post43():
    binary, version = find_motion()
    if not binary:
        return False

    return update.compare_versions(version, '4.4') >= 0  # 4.3.2 > 4.3


def has_h264_omx_support():
    binary, version, codecs = mediafiles.find_ffmpeg()
    if not binary:
        return False

    # TODO also check for motion codec parameter support

    return 'h264_omx' in codecs.get('h264', {}).get('encoders', set())


def has_h264_v4l2m2m_support():
    binary, version, codecs = mediafiles.find_ffmpeg()
    if not binary:
        return False

    # TODO also check for motion codec parameter support

    return 'h264_v4l2m2m' in codecs.get('h264', {}).get('encoders', set())


def has_h264_nvenc_support():
    binary, version, codecs = mediafiles.find_ffmpeg()
    if not binary:
        return False

    # TODO also check for motion codec parameter support

    return 'h264_nvenc' in codecs.get('h264', {}).get('encoders', set())


def has_h264_nvmpi_support():
    binary, version, codecs = mediafiles.find_ffmpeg()
    if not binary:
        return False

    # TODO also check for motion codec parameter support

    return 'h264_nvmpi' in codecs.get('h264', {}).get('encoders', set())


def has_hevc_nvmpi_support():
    binary, version, codecs = mediafiles.find_ffmpeg()
    if not binary:
        return False

    # TODO also check for motion codec parameter support

    return 'hevc_nvmpi' in codecs.get('hevc', {}).get('encoders', set())


def has_hevc_nvenc_support():
    binary, version, codecs = mediafiles.find_ffmpeg()
    if not binary:
        return False

    # TODO also check for motion codec parameter support

    return 'hevc_nvenc' in codecs.get('hevc', {}).get('encoders', set())


def has_h264_qsv_support():
    binary, version, codecs = mediafiles.find_ffmpeg()
    if not binary:
        return False

    # TODO also check for motion codec parameter support

    return 'h264_qsv' in codecs.get('h264', {}).get('encoders', set())


def has_hevc_qsv_support():
    binary, version, codecs = mediafiles.find_ffmpeg()
    if not binary:
        return False

    # TODO also check for motion codec parameter support

    return 'hevc_qsv' in codecs.get('hevc', {}).get('encoders', set())


def has_h264_nvenc_support():
    binary, version, codecs = mediafiles.find_ffmpeg()
    if not binary:
        return False

    # TODO also check for motion codec parameter support

    return 'h264_nvenc' in codecs.get('h264', {}).get('encoders', set())


def has_h264_nvmpi_support():
    binary, version, codecs = mediafiles.find_ffmpeg()
    if not binary:
        return False

    # TODO also check for motion codec parameter support

    return 'h264_nvmpi' in codecs.get('h264', {}).get('encoders', set())


def has_hevc_nvmpi_support():
    binary, version, codecs = mediafiles.find_ffmpeg()
    if not binary:
        return False

    # TODO also check for motion codec parameter support

    return 'hevc_nvmpi' in codecs.get('hevc', {}).get('encoders', set())


def has_hevc_nvenc_support():
    binary, version, codecs = mediafiles.find_ffmpeg()
    if not binary:
        return False

    # TODO also check for motion codec parameter support

    return 'hevc_nvenc' in codecs.get('hevc', {}).get('encoders', set())


def has_h264_qsv_support():
    binary, version, codecs = mediafiles.find_ffmpeg()
    if not binary:
        return False

    # TODO also check for motion codec parameter support

    return 'h264_qsv' in codecs.get('h264', {}).get('encoders', set())


def has_hevc_qsv_support():
    binary, version, codecs = mediafiles.find_ffmpeg()
    if not binary:
        return False

    # TODO also check for motion codec parameter support

    return 'hevc_qsv' in codecs.get('hevc', {}).get('encoders', set())


def resolution_is_valid(width, height):
    # width & height must be be modulo 8

    if width % 8:
        return False

    if height % 8:
        return False

    return True


async def _disable_initial_motion_detection():
    from motioneye import config

    for camera_id in config.get_camera_ids():
        camera_config = config.get_camera(camera_id)
        if not utils.is_local_motion_camera(camera_config):
            continue

        if not camera_config['@motion_detection']:
            logging.debug(
                f'motion detection disabled by config for camera with id {camera_id}'
            )
            await set_motion_detection(camera_id, False)


def _get_pid():
    motion_pid_path = os.path.join(settings.RUN_PATH, 'motion.pid')

    # read the pid from file
    try:
        with open(motion_pid_path) as f:
            return int(f.readline().strip())

    except (OSError, ValueError):
        return None



#!/usr/bin/env python3

# Copyright (c) 2022 Jean Michault
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess
import sys

import motioneye


def main():
    cmd = f"cd '{motioneye.__path__[0]}' && extra/linux_init"
    for arg in sys.argv[1:]:
        cmd += f" '{arg}'"

    subprocess.run(cmd, shell=True)


if __name__ == '__main__':
    main()



# Copyright (c) 2020 Vlsarro
# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from motioneye.handlers.base import BaseHandler

__all__ = ('LoginHandler',)


# this will only trigger the login mechanism on the client side, if required
class LoginHandler(BaseHandler):
    @BaseHandler.auth()
    def get(self):
        self.finish_json()

    def post(self):
        self.set_header('Content-Type', 'text/html')
        self.finish()



# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import logging
import os
import re
import signal
import socket
import time

import pycurl
from tornado.ioloop import IOLoop

from motioneye import config, mediafiles, meyectl, motionctl, settings, utils
from motioneye.controls import tzctl


def send_message(api_key, chat_id, message, files):
    telegram_message_url = 'https://api.telegram.org/bot%s/sendMessage' % api_key
    telegram_photo_url = 'https://api.telegram.org/bot%s/sendPhoto' % api_key
    c = pycurl.Curl()
    c.setopt(c.POST, 1)
    c.setopt(c.URL, telegram_message_url)
    if not files:
        logging.info('no files')
        c.setopt(c.POSTFIELDS, f"chat_id={chat_id}&text={message}")
        c.perform()
    else:
        logging.info('files present')
        for f in files:
            c.setopt(c.URL, telegram_photo_url)
            # Send photos
            c.setopt(
                c.HTTPPOST,
                [
                    ("chat_id", chat_id),
                    ("caption", message),
                    ("photo", (c.FORM_FILE, f)),
                ],
            )
            c.perform()
    c.close()
    logging.debug('sending telegram')


def make_message(message, camera_id, moment, timespan, callback):
    camera_config = config.get_camera(camera_id)

    # we must start the IO loop for the media list subprocess polling
    io_loop = IOLoop.current()

    def on_media_files(media_files):
        io_loop.stop()
        photos = []

        timestamp = time.mktime(moment.timetuple())
        if media_files:
            logging.debug('got media files')
            media_files = [
                m
                for m in media_files.result()
                if abs(m['timestamp'] - timestamp) < float(timespan)
            ]
            media_files.sort(key=lambda m: m['timestamp'], reverse=True)
            media_files = [
                os.path.join(camera_config['target_dir'], re.sub('^/', '', m['path']))
                for m in media_files
            ]
            logging.debug('selected %d pictures' % len(media_files))

        format_dict = {
            'camera': camera_config['camera_name'],
            'hostname': socket.gethostname(),
            'moment': moment.strftime('%Y-%m-%d %H:%M:%S'),
        }

        if settings.LOCAL_TIME_FILE:
            format_dict['timezone'] = tzctl.get_time_zone()
        else:
            format_dict['timezone'] = 'local time'

        logging.debug('creating telegram message')

        m = message % format_dict

        callback(m, media_files)

    if not timespan:
        return on_media_files([])

    logging.debug(f'waiting {float(timespan)}s for pictures to be taken')
    time.sleep(float(timespan))  # give motion some time to create motion pictures

    prefix = None
    picture_filename = camera_config.get('picture_filename')
    snapshot_filename = camera_config.get('snapshot_filename')

    if (
        (picture_filename or snapshot_filename)
        and not picture_filename
        or picture_filename.startswith('%Y-%m-%d/')
        and not snapshot_filename
        or snapshot_filename.startswith('%Y-%m-%d/')
    ):
        prefix = moment.strftime('%Y-%m-%d')
        logging.debug('narrowing down still images path lookup to %s' % prefix)

    fut = utils.cast_future(
        mediafiles.list_media(camera_config, media_type='picture', prefix=prefix)
    )
    fut.add_done_callback(on_media_files)
    io_loop.start()


def parse_options(parser, args):
    parser.description = 'Send Telegram using bot api'
    parser.add_argument('api', help='telegram api key')
    parser.add_argument('chatid', help='telegram chat room id')
    parser.add_argument('motion_camera_id', help='the id of the motion camera')
    parser.add_argument(
        'moment',
        help='the moment in ISO-8601 format',
        type=datetime.datetime.fromisoformat,
    )
    parser.add_argument('timespan', help='picture collection time span')
    return parser.parse_args(args)


def main(parser, args):
    # the motion daemon overrides SIGCHLD,
    # so we must restore it here,
    # or otherwise media listing won't work
    signal.signal(signal.SIGCHLD, signal.SIG_DFL)

    options = parse_options(parser, args)
    meyectl.configure_logging('telegram', options.log_to_file)
    logging.debug(options)
    message = 'Motion has been detected by camera "%(camera)s/%(hostname)s" at %(moment)s (%(timezone)s).'

    # do not wait too long for media list,
    # telegram notifications are critical
    settings.LIST_MEDIA_TIMEOUT = settings.LIST_MEDIA_TIMEOUT_TELEGRAM

    camera_id = motionctl.motion_camera_id_to_camera_id(options.motion_camera_id)

    def on_message(message, files):
        try:
            logging.info(f'sending telegram : {message}')
            send_message(options.api, options.chatid, message, files or [])
            logging.info('telegram sent')

        except Exception as e:
            logging.error('failed to send telegram: %s' % e, exc_info=True)

        logging.debug('bye!')

    make_message(message, camera_id, options.moment, options.timespan, on_message)



# Copyright (c) 2020 Vlsarro
# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import json
import logging
import os
import socket

from tornado.ioloop import IOLoop
from tornado.web import HTTPError

from motioneye import (
    config,
    meyectl,
    motionctl,
    remote,
    settings,
    tasks,
    template,
    uploadservices,
    utils,
)
from motioneye.controls import mmalctl, smbctl, tzctl, v4l2ctl
from motioneye.controls.powerctl import PowerControl
from motioneye.handlers.base import BaseHandler
from motioneye.utils.mjpeg import test_mjpeg_url
from motioneye.utils.rtmp import test_rtmp_url
from motioneye.utils.rtsp import test_rtsp_url

__all__ = ('ConfigHandler',)


class ConfigHandler(BaseHandler):
    async def get(self, camera_id=None, op=None):
        config.invalidate_monitor_commands()

        if camera_id is not None:
            camera_id = int(camera_id)

        if op == 'get':
            await self.get_config(camera_id)
            return

        elif op == 'list':
            await self.list()
            return

        elif op == 'backup':
            return self.backup()

        elif op == 'authorize':
            return self.authorize(camera_id)

        else:
            raise HTTPError(400, 'unknown operation')

    async def post(self, camera_id=None, op=None):
        if camera_id is not None:
            camera_id = int(camera_id)

        if op == 'set':
            await self.set_config(camera_id)
            return

        elif op == 'add':
            await self.add_camera()
            return

        elif op == 'rem':
            return self.rem_camera(camera_id)

        elif op == 'restore':
            return self.restore()

        elif op == 'test':
            await self.test(camera_id)

        else:
            raise HTTPError(400, 'unknown operation')

    @BaseHandler.auth(admin=True)
    async def get_config(self, camera_id):
        if camera_id:
            logging.debug(f'getting config for camera {camera_id}')

            if camera_id not in config.get_camera_ids():
                raise HTTPError(404, 'no such camera')

            local_config = config.get_camera(camera_id)
            if utils.is_local_motion_camera(local_config):
                ui_config = config.motion_camera_dict_to_ui(local_config)

                return self.finish_json(ui_config)

            elif utils.is_remote_camera(local_config):
                resp = await remote.get_config(local_config)
                if resp.error:
                    msg = 'Failed to get remote camera configuration for {url}: {msg}.'.format(
                        url=remote.pretty_camera_url(local_config), msg=resp.error
                    )
                    return self.finish_json_with_error(msg)

                for key, value in list(local_config.items()):
                    resp.remote_ui_config[key.replace('@', '')] = value

                # replace the real device url with the remote camera path
                resp.remote_ui_config['device_url'] = remote.pretty_camera_url(
                    local_config
                )
                return self.finish_json(resp.remote_ui_config)

            else:  # assuming simple mjpeg camera
                ui_config = config.simple_mjpeg_camera_dict_to_ui(local_config)

                return self.finish_json(ui_config)

        else:
            logging.debug('getting main config')

            ui_config = config.main_dict_to_ui(config.get_main())
            return self.finish_json(ui_config)

    @BaseHandler.auth(admin=True)
    async def set_config(self, camera_id):
        try:
            ui_config = json.loads(self.request.body)

        except Exception as e:
            logging.error(f'could not decode json: {str(e)}')

            raise

        camera_ids = config.get_camera_ids()

        async def set_camera_config(camera_id, ui_config, on_finish):
            logging.debug(f'setting config for camera {camera_id}...')

            if camera_id not in camera_ids:
                raise HTTPError(404, 'no such camera')

            local_config = config.get_camera(camera_id)
            if utils.is_local_motion_camera(local_config):
                local_config = config.motion_camera_ui_to_dict(ui_config, local_config)

                config.set_camera(camera_id, local_config)

                on_finish(None, True)  # (no error, motion needs restart)

            elif utils.is_remote_camera(local_config):
                # update the camera locally
                local_config['@enabled'] = ui_config['enabled']
                config.set_camera(camera_id, local_config)

                if 'name' in ui_config:

                    def on_finish_wrapper(e=None):
                        return on_finish(e, False)

                    ui_config['enabled'] = True  # never disable the camera remotely
                    result = await remote.set_config(local_config, ui_config)
                    return on_finish(result, False)

                else:
                    # when the ui config supplied has only the enabled state
                    # and no useful fields (such as "name"),
                    # the camera was probably disabled due to errors
                    on_finish(None, False)

            else:  # assuming simple mjpeg camera
                local_config = config.simple_mjpeg_camera_ui_to_dict(
                    ui_config, local_config
                )

                config.set_camera(camera_id, local_config)

                on_finish(None, False)  # (no error, motion doesn't need restart)

        def set_main_config(ui_config):
            logging.debug('setting main config...')

            old_main_config = config.get_main()
            old_admin_username = old_main_config.get('@admin_username')
            old_normal_username = old_main_config.get('@normal_username')
            old_lang = old_main_config.get('@lang')

            main_config = config.main_ui_to_dict(ui_config)
            main_config.setdefault('camera', old_main_config.get('camera', []))

            admin_username = main_config.get('@admin_username')
            admin_password = main_config.get('@admin_password')

            normal_username = main_config.get('@normal_username')
            normal_password = main_config.get('@normal_password')
            lang = main_config.get('@lang')

            additional_configs = config.get_additional_structure(camera=False)[1]
            reboot_config_names = [
                ('@_' + c['name'])
                for c in list(additional_configs.values())
                if c.get('reboot')
            ]
            reboot = bool(
                [
                    k
                    for k in reboot_config_names
                    if old_main_config.get(k) != main_config.get(k)
                ]
            )

            config.set_main(main_config)

            reload = False
            restart = False

            if lang != old_lang:
                logging.debug('lang changed, reload needed')
                meyectl.load_l10n()
                template._reload_lang()

                reload = True

            if admin_username != old_admin_username or admin_password is not None:
                logging.debug('admin credentials changed, reload needed')

                reload = True

            if normal_username != old_normal_username or normal_password is not None:
                logging.debug(
                    'surveillance credentials changed, all camera configs must be updated'
                )

                # reconfigure all local cameras to update the stream authentication options
                for camera_id in config.get_camera_ids():
                    local_config = config.get_camera(camera_id)
                    if not utils.is_local_motion_camera(local_config):
                        continue

                    ui_config = config.motion_camera_dict_to_ui(local_config)
                    local_config = config.motion_camera_ui_to_dict(
                        ui_config, local_config
                    )

                    config.set_camera(camera_id, local_config)

                    restart = True

            if reboot and settings.ENABLE_REBOOT:
                logging.debug('system settings changed, reboot needed')

            else:
                reboot = False

            return {'reload': reload, 'reboot': reboot, 'restart': restart}

        reload = False  # indicates that browser should reload the page
        reboot = [False]  # indicates that the server will reboot immediately
        restart = [
            False
        ]  # indicates that the local motion instance was modified and needs to be restarted
        error = [None]

        def finish():
            if reboot[0]:
                if settings.ENABLE_REBOOT:

                    def call_reboot():
                        PowerControl.reboot()

                    io_loop = IOLoop.current()
                    io_loop.add_timeout(datetime.timedelta(seconds=2), call_reboot)
                    return self.finish({'reload': False, 'reboot': True, 'error': None})

                else:
                    reboot[0] = False

            if restart[0]:
                logging.debug('motion needs to be restarted')

                motionctl.stop()

                if settings.SMB_SHARES:
                    logging.debug('updating SMB mounts')
                    stop, start = smbctl.update_mounts()  # @UnusedVariable

                    if start:
                        motionctl.start()

                else:
                    motionctl.start()

            self.finish({'reload': reload, 'reboot': reboot[0], 'error': error[0]})

        if camera_id is not None:
            if camera_id == 0:  # multiple camera configs
                if len(ui_config) > 1:
                    logging.debug('setting multiple configs')

                elif len(ui_config) == 0:
                    logging.warning('no configuration to set')

                    return self.finish()

                so_far = [0]

                def check_finished(e, r):
                    restart[0] = restart[0] or r
                    error[0] = error[0] or e
                    so_far[0] += 1

                    if so_far[0] >= len(ui_config):  # finished
                        finish()

                # make sure main config is handled first
                items = list(ui_config.items())
                items.sort(key=lambda key_cfg: key_cfg[0] != 'main')

                for key, cfg in items:
                    if key == 'main':
                        result = set_main_config(cfg)
                        reload = result['reload'] or reload
                        reboot[0] = result['reboot'] or reboot[0]
                        restart[0] = result['restart'] or restart[0]
                        check_finished(None, False)

                    else:
                        await set_camera_config(int(key), cfg, check_finished)

            else:  # single camera config

                def on_finish(e, r):
                    error[0] = e
                    restart[0] = r
                    finish()

                await set_camera_config(camera_id, ui_config, on_finish)

        else:  # main config
            result = set_main_config(ui_config)
            reload = result['reload']
            reboot[0] = result['reboot']
            restart[0] = result['restart']

    def _handle_list_cameras_response(self, resp: utils.GetCamerasResponse):
        if resp.error:
            return self.finish_json_with_error(resp.error)
        else:
            return self.finish_json({'cameras': resp.cameras})

    def finish_json_with_error(self, error_msg: str):
        return self.finish_json({'error': error_msg})

    def check_finished(self, cameras: list, length: list) -> bool:
        if len(cameras) == length[0]:
            cameras.sort(key=lambda c: c['id'])
            self.finish_json({'cameras': cameras})
            return True
        else:
            return False

    def _handle_get_config_response(
        self,
        camera_id,
        local_config,
        resp: utils.GetConfigResponse,
        cameras: list,
        length: list,
    ) -> None:
        if resp.error:
            cameras.append(
                {
                    'id': camera_id,
                    'name': '&lt;' + remote.pretty_camera_url(local_config) + '&gt;',
                    'enabled': False,
                    'streaming_framerate': 1,
                    'framerate': 1,
                }
            )

        else:
            resp.remote_ui_config['id'] = camera_id

            if not resp.remote_ui_config['enabled'] and local_config['@enabled']:
                # if a remote camera is disabled, make sure it's disabled locally as well
                local_config['@enabled'] = False
                config.set_camera(camera_id, local_config)

            elif resp.remote_ui_config['enabled'] and not local_config['@enabled']:
                # if a remote camera is locally disabled, make sure the remote config says the same thing
                resp.remote_ui_config['enabled'] = False

            for key, value in list(local_config.items()):
                resp.remote_ui_config[key.replace('@', '')] = value

            cameras.append(resp.remote_ui_config)

        return self.check_finished(cameras, length)

    @BaseHandler.auth()
    async def list(self):
        logging.debug('listing cameras')

        proto = self.get_argument('proto')
        if proto == 'motioneye':  # remote listing
            return self._handle_list_cameras_response(
                await remote.list_cameras(self.get_all_arguments())
            )

        elif proto == 'netcam':
            scheme = self.get_argument('scheme', 'http')

            if scheme in ['http', 'https', 'mjpeg']:
                resp = await test_mjpeg_url(
                    self.get_all_arguments(), auth_modes=['basic'], allow_jpeg=True
                )
                return self._handle_list_cameras_response(resp)

            elif scheme == 'rtsp':
                resp = await test_rtsp_url(self.get_all_arguments())
                return self._handle_list_cameras_response(resp)

            elif scheme == 'rtmp':
                resp = test_rtmp_url(self.get_all_arguments())
                return self._handle_list_cameras_response(resp)

            else:
                return self.finish_json_with_error(f'protocol {scheme} not supported')

        elif proto == 'mjpeg':
            resp = await test_mjpeg_url(
                self.get_all_arguments(),
                auth_modes=['basic', 'digest'],
                allow_jpeg=False,
            )
            return self._handle_list_cameras_response(resp)

        elif proto == 'v4l2':
            configured_devices = set()
            for camera_id in config.get_camera_ids():
                data = config.get_camera(camera_id)
                if utils.is_v4l2_camera(data):
                    configured_devices.add(data['videodevice'])

            cameras = [
                {'id': d[1], 'name': d[2]}
                for d in v4l2ctl.list_devices()
                if (d[0] not in configured_devices) and (d[1] not in configured_devices)
            ]

            return self.finish_json({'cameras': cameras})

        elif proto == 'mmal':
            configured_devices = set()
            for camera_id in config.get_camera_ids():
                data = config.get_camera(camera_id)
                if utils.is_mmal_camera(data):
                    configured_devices.add(data['mmalcam_name'])

            cameras = [
                {'id': d[0], 'name': d[1]}
                for d in mmalctl.list_devices()
                if (d[0] not in configured_devices)
            ]

            return self.finish_json({'cameras': cameras})

        else:  # assuming local motionEye camera listing
            cameras = []
            camera_ids = config.get_camera_ids()
            if not config.get_main().get('@enabled'):
                camera_ids = []

            length = [len(camera_ids)]

            for camera_id in camera_ids:
                local_config = config.get_camera(camera_id)
                if local_config is None:
                    continue

                if utils.is_local_motion_camera(local_config):
                    ui_config = config.motion_camera_dict_to_ui(local_config)
                    cameras.append(ui_config)
                    if self.check_finished(cameras, length):
                        return

                elif utils.is_remote_camera(local_config):
                    if (
                        local_config.get('@enabled')
                        or self.get_argument('force', None) == 'true'
                    ):
                        resp = await remote.get_config(local_config)
                        if self._handle_get_config_response(
                            camera_id, local_config, resp, cameras, length
                        ):
                            return

                    else:  # don't try to reach the remote of the camera is disabled
                        if self._handle_get_config_response(
                            camera_id,
                            local_config,
                            utils.GetConfigResponse(None, error=True),
                            cameras,
                            length,
                        ):
                            return

                else:  # assuming simple mjpeg camera
                    ui_config = config.simple_mjpeg_camera_dict_to_ui(local_config)
                    cameras.append(ui_config)
                    if self.check_finished(cameras, length):
                        return

            return self.finish_json({'cameras': cameras})

    @BaseHandler.auth(admin=True)
    async def add_camera(self):
        logging.debug('adding new camera')

        try:
            device_details = json.loads(self.request.body)

        except Exception as e:
            logging.error(f'could not decode json: {str(e)}')

            raise

        camera_config = config.add_camera(device_details)

        if utils.is_local_motion_camera(camera_config):
            motionctl.stop()

            if settings.SMB_SHARES:
                stop, start = smbctl.update_mounts()  # @UnusedVariable

                if start:
                    motionctl.start()

            else:
                motionctl.start()

            ui_config = config.motion_camera_dict_to_ui(camera_config)

            return self.finish_json(ui_config)

        elif utils.is_remote_camera(camera_config):
            resp = await remote.get_config(camera_config)
            if resp.error:
                return self.finish_json_with_error(resp.error)

            for key, value in list(camera_config.items()):
                resp.remote_ui_config[key.replace('@', '')] = value

            return self.finish_json(resp.remote_ui_config)

        else:  # assuming simple mjpeg camera
            ui_config = config.simple_mjpeg_camera_dict_to_ui(camera_config)

            return self.finish_json(ui_config)

    @BaseHandler.auth(admin=True)
    def rem_camera(self, camera_id):
        logging.debug(f'removing camera {camera_id}')

        local = utils.is_local_motion_camera(config.get_camera(camera_id))
        config.rem_camera(camera_id)

        if local:
            motionctl.stop()
            motionctl.start()

        return self.finish_json()

    @BaseHandler.auth(admin=True)
    def backup(self):
        content = config.backup()

        if not content:
            raise Exception('failed to create backup file')

        filename = 'motioneye-config.tar.gz'
        self.set_header('Content-Type', 'application/x-compressed')
        self.set_header('Content-Disposition', 'attachment; filename=' + filename + ';')

        return self.finish(content)

    @BaseHandler.auth(admin=True)
    def restore(self):
        try:
            content = self.request.files['files'][0]['body']

        except KeyError:
            raise HTTPError(400, 'file attachment required')

        result = config.restore(content)
        if result:
            return self.finish_json({'ok': True, 'reboot': result['reboot']})

        else:
            return self.finish_json({'ok': False})

    @classmethod
    def _on_test_result(cls, result):
        upload_service_test_info = getattr(cls, '_upload_service_test_info', None)
        cls._upload_service_test_info = None

        if not upload_service_test_info:
            return logging.warning('no pending upload service test request')

        (request_handler, service_name) = upload_service_test_info

        if result is True:
            logging.info(f'accessing {service_name} succeeded.result {result}')
            return request_handler.finish_json()

        else:
            logging.warning(f'accessing {service_name} failed: {result}')
            return request_handler.finish_json({'error': result})

    @BaseHandler.auth(admin=True)
    async def test(self, camera_id):
        what = self.get_argument('what')
        data = self.get_all_arguments()
        camera_config = config.get_camera(camera_id)

        if utils.is_local_motion_camera(camera_config):
            if what == 'upload_service':
                service_name = data['service']
                ConfigHandler._upload_service_test_info = (self, service_name)

                result = uploadservices.test_access(
                    camera_id=camera_id, service_name=service_name, data=data
                )
                logging.debug(f'test access {service_name} result {result}')
                if result is True:
                    logging.info(f'accessing {service_name} succeeded.result {result}')
                    return self.finish_json()
                else:
                    logging.warning(f'accessing {service_name} failed: {result}')
                    return self.finish_json({'error': result})

            elif what == 'email':
                import smtplib

                from motioneye import sendmail

                logging.debug('testing notification email')

                try:
                    subject = sendmail.subjects['motion_start']
                    message = sendmail.messages['motion_start']
                    format_dict = {
                        'camera': camera_config['camera_name'],
                        'hostname': socket.gethostname(),
                        'moment': datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                    }
                    if settings.LOCAL_TIME_FILE:
                        format_dict['timezone'] = tzctl.get_time_zone()

                    else:
                        format_dict['timezone'] = 'local time'

                    message = message % format_dict
                    subject = subject % format_dict

                    old_timeout = settings.SMTP_TIMEOUT
                    settings.SMTP_TIMEOUT = 10
                    sendmail.send_mail(
                        data['smtp_server'],
                        int(data['smtp_port']),
                        data['smtp_account'],
                        data['smtp_password'],
                        data['smtp_tls'],
                        data['from'],
                        [data['addresses']],
                        subject=subject,
                        message=message,
                        files=[],
                    )

                    settings.SMTP_TIMEOUT = old_timeout
                    logging.debug('notification email test succeeded')
                    return self.finish_json()

                except Exception as e:
                    if isinstance(e, smtplib.SMTPResponseException):
                        msg = e.smtp_error

                    else:
                        msg = str(e)

                    msg_lower = msg.lower()
                    if msg_lower.count('tls'):
                        msg = 'TLS might be required'

                    elif msg_lower.count('authentication'):
                        msg = 'authentication error'

                    elif msg_lower.count('name or service not known'):
                        msg = 'check SMTP server name'

                    elif msg_lower.count('connection refused'):
                        msg = 'check SMTP port'

                    logging.error(
                        'notification email test failed: %s' % msg, exc_info=True
                    )
                    return self.finish_json({'error': str(msg)})

            elif what == 'telegram':
                from motioneye import sendtelegram

                logging.debug('testing telegram notification')

                try:
                    message = 'This is a test of motionEye\'s telegram messaging'
                    sendtelegram.send_message(
                        data['api'], int(data['chatid']), message=message, files=[]
                    )

                    self.finish_json()

                    logging.debug('telegram notification test succeeded')

                except Exception as e:
                    msg = str(e)

                    msg_lower = msg.lower()
                    logging.error(
                        'telegram notification test failed: %s' % msg, exc_info=True
                    )
                    self.finish_json({'error': str(msg)})

            elif what == 'network_share':
                logging.debug(
                    'testing access to network share //{}/{}'.format(
                        data['server'], data['share']
                    )
                )

                try:
                    smbctl.test_share(
                        data['server'],
                        data['share'],
                        data['smb_ver'],
                        data['username'],
                        data['password'],
                        data['root_directory'],
                    )

                    logging.debug(
                        'access to network share //{}/{} succeeded'.format(
                            data['server'], data['share']
                        )
                    )
                    return self.finish_json()

                except Exception as e:
                    logging.error(
                        'access to network share //{}/{} failed: {}'.format(
                            data['server'], data['share'], e
                        )
                    )
                    return self.finish_json({'error': str(e)})

            else:
                raise HTTPError(400, 'unknown test %s' % what)

        elif utils.is_remote_camera(camera_config):
            resp = await remote.test(camera_config, data)
            if resp.result is True:
                return self.finish_json()
            else:
                result = resp.result or resp.error
                return self.finish_json_with_error(result)

        else:
            raise HTTPError(400, 'cannot test features on this type of camera')

    @BaseHandler.auth(admin=True)
    def authorize(self, camera_id):
        service_name = self.get_argument('service')
        if not service_name:
            raise HTTPError(400, 'service_name required')

        url = uploadservices.get_authorize_url(service_name)
        if not url:
            raise HTTPError(
                400, 'no authorization url for upload service %s' % service_name
            )

        logging.debug('redirected to authorization url %s' % url)
        self.redirect(url)
